-----------------------------------------------------------------------------------------------
-- Client Lua Script for ForgottenChat
-----------------------------------------------------------------------------------------------
 
require "Window"
 
local ForgottenChat = {}
-----------------------------------------------------------------------------------------------
-- Utility Functions and Constants
-----------------------------------------------------------------------------------------------
function split(strInput, separator)
	local returnSegments={}
	for i in string.gmatch(strInput, separator) do
	  table.insert(returnSegments, i)
	end
	return returnSegments
end
function trim(s)
  -- from PiL2 20.4
  return (s:gsub("^%s*(.-)%s*$", "%1"))
end
function isBefore(dateA, dateB)
	--Checks to see if dateA is before dateB
	if dateA.nYear<dateB.nYear then
		return true
	elseif dateA.nYear>dateB.nYear then
		return false
	end
	--From this point onward, we can assume the two dates fall within the same year
	if dateA.nMonth<dateB.nMonth then
		return true
	elseif dateA.nMonth>dateB.nMonth then
		return false
	end
	--From this point onward, we can assume the two dates fall within the same month and year
	if dateA.nDay<dateB.nDay then
		return true
	elseif dateA.nDay>dateB.nDay then
		return false
	end
	--From this point onward, we can assume the two dates fall within the same day, month and year
	if dateA.nHour<dateB.nHour then
		return true
	elseif dateA.nHour>dateB.nHour then
		return false
	end
	--From this point onward, we can assume the two dates fall within the same hour, day, month and year
	if dateA.nMinute<dateB.nMinute then
		return true
	elseif dateA.nMinute>dateB.nMinute then
		return false
	end
	--From this point onward, we can assume the two dates fall within the same minute, hour, day, month and year
	if dateA.nSecond<dateB.nSecond then
		return true
	elseif dateA.nSecond>=dateB.nSecond then
		return false
	end
end
function FindDescendentByName(sourceFrame, descendentName)
    --This function performs a recursive, depth-first search of all descendents of a xml node for
    --the first node with the specified name, and returns it
    if sourceFrame:GetName()==descendentName then
        return sourceFrame
    else
        for index, child in pairs(sourceFrame:GetChildren()) do
            local testChild=FindDescendentByName(child, descendentName) 
            if testChild~=nil then return testChild end
        end
    end
end

local daysOfTheWeek = {"Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"}
local monthsOfTheYear = {"January","February","March","April","May","June","July","August","September","October","November","December"}
local emoticonLookup = {
    --Angry
    [">:("]="ForgottenChat:Angry",
    [">:-("]="ForgottenChat:Angry",
    --Embarrased
    [":s"]="ForgottenChat:Embarrased",
    [":S"]="ForgottenChat:Embarrased",
    [":$"]="ForgottenChat:Embarrased",
    [":-s"]="ForgottenChat:Embarrased",
    [":-S"]="ForgottenChat:Embarrased",
    [":-$"]="ForgottenChat:Embarrased",
    --Emotionless
    [":|"]="ForgottenChat:Emotionless",
    [":-|"]="ForgottenChat:Emotionless",
    --Happy
    [":)"]="ForgottenChat:Happy",
    [":-)"]="ForgottenChat:Happy",
    [":-]"]="ForgottenChat:Happy",
    [":]"]="ForgottenChat:Happy",
    ["=]"]="ForgottenChat:Happy",
    ["=)"]="ForgottenChat:Happy",
    --Joy
    [":')"]="ForgottenChat:Joy",
    [":'-)"]="ForgottenChat:Joy",
    [":'-]"]="ForgottenChat:Joy",
    [":']"]="ForgottenChat:Joy",
    ["=']"]="ForgottenChat:Joy",
    ["=')"]="ForgottenChat:Joy",
    --Sad
    [":("]="ForgottenChat:Sad",
    [":-("]="ForgottenChat:Sad",
    [":-["]="ForgottenChat:Sad",
    [":["]="ForgottenChat:Sad",
    ["=("]="ForgottenChat:Sad",
    ["=["]="ForgottenChat:Sad",
    --Silly
    [":p"]="ForgottenChat:Silly",
    [":-p"]="ForgottenChat:Silly",
    [":P"]="ForgottenChat:Silly",
    [":-{"]="ForgottenChat:Silly",
    --Tear
    [":'("]="ForgottenChat:Tear",
    [":'-("]="ForgottenChat:Tear",
    [":'-["]="ForgottenChat:Tear",
    [":'["]="ForgottenChat:Tear",
    ["='["]="ForgottenChat:Tear",
    ["='["]="ForgottenChat:Tear",
    --VeryHappy
    [":D"]="ForgottenChat:VeryHappy",
    [":-D"]="ForgottenChat:VeryHappy",
    [":))"]="ForgottenChat:VeryHappy",
    ["=D"]="ForgottenChat:VeryHappy",
    ["=-D"]="ForgottenChat:VeryHappy",
    --Wink
    [";)"]="ForgottenChat:Wink",
    [";-)"]="ForgottenChat:Wink",
    [";-]"]="ForgottenChat:Wink",
    [";]"]="ForgottenChat:Wink",
}
local karEvalColors =
{
	[Item.CodeEnumItemQuality.Inferior] 		= "ItemQuality_Inferior",
	[Item.CodeEnumItemQuality.Average] 			= "ItemQuality_Average",
	[Item.CodeEnumItemQuality.Good] 			= "ItemQuality_Good",
	[Item.CodeEnumItemQuality.Excellent] 		= "ItemQuality_Excellent",
	[Item.CodeEnumItemQuality.Superb] 			= "ItemQuality_Superb",
	[Item.CodeEnumItemQuality.Legendary] 		= "ItemQuality_Legendary",
	[Item.CodeEnumItemQuality.Artifact]		 	= "ItemQuality_Artifact",
}
local reverseClassEnum = {[0]="Classy Guy"}
for class, code in pairs(GameLib.CodeEnumClass) do
    reverseClassEnum[code]=class
end
local reverseRaceEnum = {}
for race, code in pairs(GameLib.CodeEnumRace) do
    reverseRaceEnum[code]=race
end
local reversePathEnum = {[0]="Soldier", "Settler", "Scientist", "Explorer"}
local reverseGenderEnum = {[0]="Male","Female", ""}
local reverseGenderPronounEnum = {[0]="his","her", "his"}
local reverseGenderPronoun2Enum = {[0]="he","she", "he"}
local reverseGenderPronoun3Enum = {[0]="him","her", "him"}
local reverseGenderPronoun4Enum = {[0]="man","woman", "man"}
local reverseGenderPronoun5Enum = {[0]="boy","girl", "boy"}
local assumptionList = {
    --Pop Culture References
    "SPACESHIP?  SPACESHIP!",
    ":Genderpronoun: father smells of elderberries",
    ":Genderpronoun: mother is a hampster",
    "Lost in time and space",
    "Went back in time to 1955 and kissed :genderpronoun: mother",
    "Knock, Knock Knock... :myname:?",
    "Today :genderpronoun2: is cancelling the apocalypse!",
    "The most talented, most interesting, and most extraordinary person in the universe",
    "Did the Kessel Run in under 12 Parsecs",
    "Wants to trade :genderpronoun: very best gun for your wife",
    "Can kill you with :genderpronoun: brain",
    "Plays with plastic dinosaurs.  Grrr Rawr!",
    "In that... Special level of hell",
    "Playing Galaga! Thought you wouldn't notice, but you did",
    "Recently become an expert in thermonuclear astrophysics",
    "Repairs air conditioners as a hobby",
    "The One Who Knocks",
    ":Genderpronoun2: Who Must Not Be Named",
    "Drinks tea with The King In Yellow",
    "So ugly :genderpronoun: could be a modern art masterpiece!",
    "In a hole in the ground there lived a :race:",
    "Life before Death. Strength before Weakness. Journey before Destination.",
    "Escaped from prison using nothing but a spoon",
    "Makes soap from human fat",
    "Lives at 221b Baker Street",
    "Too old for this S#!%!",
    "Has a bad feeling about this",
    "Religious prophet from Mars",
    "Joined the Mobile Infantry and saved the world",
    "Would liked to have seen Montana",
    "Knows how that S#!% works",
    "Cocked, locked and ready to rock",
    "Gonna choke :genderpronoun: million bucks out of you!",
    "The Dragon Reborn",
    "Looking for a dweeb who wears green fatigues",
    "In a relationship with a ghost",
    ":Genderpronoun: name is :name:, you killed :genderpronoun: father",
    "The Six-Fingered :Genderpronoun4:",
    "The Gatekeeper",
    "The Keymaster",
    ":Genderpronoun: parents named the dog :name:",
    "Shipped snakes to LA by air freight",
    "Took your Swingline stapler",
    "Knows what a TV dinner feels like",
    "Would buy that for a dollar!",
    "Got dysentery on the Oregon Trail",
    "Untouchable",
    "Winner of the Ultimate Showdown",
    "Genuine Pinball Wizard",
    "Misplaced the One Ring",
    "Found the One Ring",
    "Destroyed the One Ring",
    ":Genderpronoun: old life lays behind in the mists",
    "Veteran of Serenity Valley",
    "Contemplating the meaning of life",
    "Political. :Genderpronoun2:'s cleaning up the streets",
    "Does the weird stuff",
    "Misunderstood Supervillain",
    "A hero in :genderpronoun: own way",
    ":Genderpronoun2: smells like poo!",
    "Knows what war is good for",
    "High Functioning Sociopath",
    "An insect wearing the skin of a :race:",
    "Accidentally took the Fourier Transform of :genderpronoun: cat",
    "The engineer who installed the Red Button",
    "Trapped on the surface of a sphere",
    "Who let the dogs out? :name:!",
    "Singing The Doom Song",
    "Inept Invader",
    "+++ OUT OF CHEESE ERROR.  REDO FROM START +++",
    "Never gonna give you up",
    "Never gonna let you down",
    "Never gonna run around and desert you",
    "Doublethinker",
    "Ask :genderpronoun3: to prove that 2+2=5",
    "You may call :genderpronoun: 'V'",
    "Humble vaudevillian veteran, both victim and villain",
    "Ghost In The Machine",
    "The Dread Pirate :name:",
    "The Fourth Musketeer",
    "The Fifth Horseman",
    "Speaker for the Dead",
    "Thinks 'moist' is a dirty word",
    ":Genderpronoun2: didn't bring enough grenades",
    ":Genderpronoun2: never speaks about F... Oh. Wait. Can't say that",
    "Took the blue pill",
    "Took the red pill",
    "Never send a :race: to do a Mechari's job",
    "Captain of the Night's Watch",
    "More machine than :genderpronoun4:",
    ":name:... :name:... :name:! :name:?",
    "Built a city on Rock & Roll.  Literally.",
    "Has a Sonic Screwdriver",
    "Pilots a 03-K64-Firefly Class Starship",
    "Arrogant Worm",
    "Cancer-ridden Meth Cook",
    ":Genderpronoun2: is faster than a speeding bullet",
    ":Genderpronoun2: is really the Green Lantern",
    "Oathbreaker",
    "Oathkeeper",
    "Knows what you did last summer.  Saw it on TV.",
    "Living the same day over and over again",
    "Aging backwards",
    "Rejected reality and substituted :genderpronoun: own",
    "Time Travelling Super Solider",
    "Shipwrecked on a deserted island",
    "Expert Cryptologist",
    "Too sexy for :genderpronoun: shirt!",
    "Didn't Panic",
    "Forgot :genderpronoun: towel at home",
    "Master of the Thieves Guild",
    "Thinks the world is like an onion",
    "Bends spoons with :genderpronoun: mind",
    ":name:, Memory Adept",
    ":name:, Mentor of Heroes",
    ":name:, Champion of Justice",
    ":name: The Seeker",
    "Farting in your general direction",
    "Lied about the cake",
    "Blinded by science!",
    "You know nothing :name:",
    "Trapped inside a teleporter beam",
    "EXTERMINATE!",
    "A petty thief with delusions standing?",
    "Professional Blitzball player",
    "Sentient Robotic Dinosaur",
    "Rosebud...",
    "You made :genderpronoun3: an offer :genderpronoun2: couldn't refuse",
    ":name:, in the Study, with the Spanner",
    "Tilting at windmills",
    "Worshipfulness",
    "Juggles geese for recreation",
    ":Genderpronoun2: had a good time at the D'Arbanville's ball",
    "An insult to the psychotic low-life community",
    "The key seems to be giving :name: a heavy stick and standing back",
    "Remembers Tahiti.  It's a magical place",
    ":Genderpronoun2:'s off bringing religiosity to the fuzzy-wuzzies or some such",
    "Knows what the Chain of Command is",
    "Done the impossible, and that makes :genderpronoun3: mighty",
    "Cannot abide useless people",
    "Not the plucky hero",
    ":Genderpronoun2:'ll kill a man in a fair fight... or if :Genderpronoun2: think he's gonna start a fair fight",
    ":Genderpronoun2: can't stop the signal",
    "Secrets are not :genderpronoun: concern. Keeping them is",
    "Being a bad guy",
    "Swallowed a bug",
    "Lead singer of the :name: Brothers Band",
    "Silly Walker",
    ":Genderpronoun2:'s been THUNDERSTRUCK!",
    "Patience isn't :genderpronoun: strong suit",
    "Asleep for the action and awake for the money, as usual",
    ":Genderpronoun2: is :name:",
    "The fate of 12 billion people is in :genderpronoun: hands",
    "Going to die surrounded by the biggest idiots in the galaxy",
    "Metaphors go over :genderpronoun: head",
    "Ain't no thing like :genderpronoun3:, except :genderpronoun3:!",
    "The Real Slim :name:",
    ":Genderpronoun3: coat is kinda brown...ish",
    "Carrying on as if nothing really matters",
    "Just a poor :genderpronoun5:, :genderpronoun2: needs no sympathy", 
    "Ice Farmer with a pet elk",
    "Suiting up!",
    "Playing chess with the devil",
    "Trapped on Isla Sorna",
    "PC LOAD LETTER",
    "Child of Night",
    "Doctors hate :genderpronoun3:",
    "Blonde King-Child",
    
    --Wildstar Self-Referencing
    "Owns a cubig farm",
    "Writes sappy love poetry in Eldanen",
    "Living the Nexus Dream",
    "Artemis Zin in disguise",
    "Teenage Mutant Ninja Jabbit",
    "Slayer of Chompacabras",
    
    --Titles/Adjectives
    "Adopted",
    "The Prophesied One",
    "A God Among Men",
    "Super Secret Spy",
    "Super, SUPER Secret Spy",
    "Exiled President of a third world country",
    "Genius",
    "One of a kind",
    "High School Heart Throb",
    "Leader of a three ring circus",
    "Pyromaniac",
    "Spendthrift",
    "Philanthropist",
    "Kingmaker",
    "Nobody of consequence",
    "Army of One",
    "Bigger.  Stronger.  Faster",
    "One Shot.  One Kill",
    "Fashionista",
    "Existentialist",
    "Vegetarian",
    "Astronaut",
    "World Famous",
    "Callipygian",
    "Voracious Sesquipedalian",
    "Cool Dude",
    "Programmer Extraordinaire!",
    "Forgetful",
    "Brewmaster",
    "Postmaster General",
    "Peaceful protestor",
    "Stand-up Comedian",
    "Last known descendant of royalty",
    "Historical Fiction author",
    "Nosepicker",
    "Sultan of Swing",
    "Fawning fool-born fustilarian",
    "Foot Fetishist",
    "More Human than :race:",
    "Master Assassin",
    "Thousand Year Old Vampire",
    "Supreme Emperor of all Mankind",
    "Grand Poobah",
    ":name:, Destroyer of Worlds",
    ":name:, Destroyer of Fuzzy Animals",
    "Mad Scientist",
    "Shogun",
    "Genetically perfect",
    "President of the :myname: Fan Club",
    "The Comedic Relief",
    "Omnipotent",
    
    --History, Skill or Personality
    "The epitome of cool",
    "Speaks in tongues",
    "Dresses in drag",
    "Beat up Chuck Norris",
    "Killed a man once.  You're next",
    "Has a bullet with your name on it",
    "Slipped while shaving this morning.  Needs surgery",
    "Recently gave up :genderpronoun: diet",
    "Collects fine china",
    "That really enthousiastic guy",
    "Has an inferiority complex",
    "Not you.  Isn't :genderpronoun2: lucky?",
    "Stalking you from the bushes out front",
    "Destroyed the world, but put it back together again",
    ":Genderpronoun2:'s got your back!",
    "Writing :genderpronoun: life story",
    "Hates being right all the time",
    "Picked a hell of a day to quit drinking",
    "Has a pistol duel at dawn",
    "Has a mean golf swing",
    "Sold into slavery for 13c",
    "Rolled 21 on a D20 once",
    "Has a secret room in :genderpronoun: basement",
    "Got :genderpronoun: brains sucked out",
    "Eats bacon for breakfast, lunch and dinner",
    "Has done questionable things",
    "Has done extraordinary things",
    "Has done disturbing things",
    "Planning a bank heist",
    "Likes long walks on the beach",
    "Has big feet",
    "Doesn't like :genderpronoun: siblings",
    "Threw a party and nobody came",
    "Living in a Shakespearean comedy",
    "Enjoys dancing in the rain",
    "Treating :genderpronoun3:self to a chocolate bar",
    "Knows when :genderpronoun2: will die",
    "OMG IS IT REALLY :name:??!?",
    "Afraid of loud noises",
    "Afraid of :genderpronoun: own dog",
    "Contemplating the universe",
    "You'd date :genderpronoun3:",
    "Won the lottery.  Twice.",
    "Sings to :genderpronoun3:self in the shower",
    "Talks to :genderpronoun3:self when nobody is looking",
    "Cured cancer.  Too drunk to remember it.",
    "Declared war on :genderpronoun: furniture",
    "Made peace with the potted plants",
    "Lives in a trash heap, and proud of it!",
    "Has a restraining order against you",
    "You have a restraining order against :genderpronoun3:",
    "Literally rolling on the floor laughing",
    "Thinks the world is flat",
    "Knows the world really is flat",
    "Thinks :genderpronoun2: is an onion",
    ":Genderpronoun: cat is on the keyboard",
    "Speaks Esperanto",
    "Taking cooking lessons",
    "Thinking about :genderpronoun: glutes",
    "Opens beer bottles with :genderpronoun: eyeball",
    "Loves everyone unconditionally",
    "Eats vegetarians for breakfast",
    "Playing 'The Game'",
    "Thinks :genderpronoun2: got away with it",
    "Collects children's toys",
    "Has secrets nobody will ever find out",
    "Ssshhhh... :genderpronoun2:'s sleeping",
    "Asleep.  :Genderpronoun2:'s dreaming about monkeys",
    "Recently taken up practicing voodoo",
    "Took candy from a child, but felt bad about it",
    ":Genderpronoun2:'s trustworthy.  Probably.",
    "Has a paper due in the morning that :genderpronoun2: forgot about",
    "Never learned to swim.  Afraid to admit it",
    "Gave to charity this morning",
    "The Pope's got nothing on :genderpronoun3:",
    "Writes folk songs",
    "Accidentally shot :genderpronoun: own foot",
    "Busy solving P=NP",
    "Knows the solution to P=NP",
    "Has 'BEAT DOWN' finger tats",
    "Rescues cats from trees",
    "Makes the BEST chocolate chip cookies EVER!",
    "Aint the sharpest tool in the shed",
    "Thinks 'Defenestration' means something dirty",
    "Your best friend",
    "Your arch nemesis",
    ":Genderpronoun2: has two dads",
    ":Genderpronoun2: has two moms",
    "Was 'King of the World' in one of :genderpronoun: previous lives",
    "Has anger issues.  Be careful",
    "Can't remember anything more than 11.45 seconds ago",
    ":Genderpronoun2:'s planning the perfect crime",
    ":Genderpronoun2: will end the world in 9 days",
    ":Genderpronoun2: once lit :genderpronoun: own hair on fire",
    ":Genderpronoun2: thinks the whole world is one sick joke",
    "Can fart the alphabet if prompted",
    "Doesn't know who :genderpronoun2: is",
    "Drinks :genderpronoun: beer warm",
    "Invented sliced bread",
    "Invented unsliced bread.  Still kicking himself",
    "The One.  The Only! The :name:!",
    "Hiding from the collection agency",
    "Drinks straight from the bottle",
    "Hacked the NSA for laughs",
    "Trapped in a nine-dimensional bubble",
    "Imperfect Perfectionist",
    "Possessed by a Malevolent Leprechaun",
    "Driven to succeed",
    "Blissfully ignorant",
    "Willfully ignorant",
    "Steaming at the ears",
    "Everywhere at once",
    "A figment of your imagination",
    "Uses too many exlamation points!!!11!one!",
    "Telipna Albofasciata Chaotica",
    "Fanatical about :genderpronoun: hygiene",
    "Half :genderpronoun4:, half spider",
    "Collects antique paperweights",
    "The greatest beard you've ever seen on a :genderpronoun4:",
    "Wishes :genderpronoun2: was a Velociraptor",
    
    --Other
    "AAAAAAAAAAAAAAAAAAH!",
    "Grrrrr.  Arrg",
    "1337",
    "Try '/fc 50now!'",
    "010010000110010101101100011011000110111100100001",
    
    --Unsorted
    "Professional Spaceship Racer",
    "Underappreciated Lyricist",
    "Self-replicating nanobot",
    "Degenerate gambler",
    "Unpaid Intern",
    "Epic personality",
    "Brilliant",
    "Likes the taste of pennies",
    "Prefers water in :genderpronoun: cereal",
    "Eats nothing but magenta foods",
    "Thinks :genderpronoun2:'s 'All That'",
    "Chaos Theorist",
    "Confused by flowcharts",
    "Confounded by rudamentary Quantum Physics",
    "Confused by Venn Diagrams",
    "Huge Amorphous Gelatinous Entity",
    "Hivemind",
    "Invisible",
    ":Genderpronoun: shoes are too tight",
    "Trapped between a rock and a hard place",
    "The Forgotten Lord of Nexus",
    "Coming at you bro",
    "Slowly going crazy is :genderpronoun2: 1, 2, 3, 4, 5, 6, switch",
    "Chill.  Totally Chill",
    "Fragging it up in high style",
    "Hoverboard mechanic",
    "Never forgets a face",
    "Just woke up",
    "Recovering from a gut wound",
    ":Genderpronoun2: once ate a wooden chair on a dare",
    ":Genderpronoun2:'s a Poet, and didn't even know it",
    
    
}

--local reverseGenderPronounEnum = {[0]="his","her", "his"}
--local reverseGenderPronoun2Enum = {[0]="he","she", "he"}
--local reverseGenderPronoun3Enum = {[0]="him","her", "him"}
--local reverseGenderPronoun4Enum = {[0]="man","woman", "man"}
--local reverseGenderPronoun5Enum = {[0]="boy","girl", "boy"}
-----------------------------------------------------------------------------------------------
-- ForgottenChat Initialization functions
-----------------------------------------------------------------------------------------------
function ForgottenChat:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	self.conversations={}
	self.mostRecentIncomingConversation=nil    
    self.skins={
        ["Default"]={
            ["enabled"]=true,
            ["suffix"]="Default",
            ["name"]="Largesse (Default)",
            ["minimize"]="normal",
        },
        ["Minimalist"]={
            ["enabled"]=true,
            ["suffix"]="Minimalist",
            ["name"]="Slim Jim",
            ["minimize"]="normal",
        },
        ["Holo"]={
            ["enabled"]=true,
            ["suffix"]="Holo",
            ["name"]="Holography",
            ["minimize"]="normal",
        },
        ["Blood"]={
            ["enabled"]=true,
            ["suffix"]="Blood",
            ["name"]="Crimson Claxon",
            ["minimize"]="normal",
        },
        ["Gold"]={
            ["enabled"]=true,
            ["suffix"]="Gold",
            ["name"]="All That Glitters",
            ["minimize"]="normal",
        },
        ["Hexes"]={
            ["enabled"]=true,
            ["suffix"]="Hex",
            ["name"]="As Yet Unnamed",
            ["minimize"]="normal",
        },
        ["Warriotox"]={
            ["enabled"]=true,
            ["suffix"]="Tox",
            ["name"]="Codename: Warriotox",
            ["minimize"]="reverse",
        },
        ["Jos1"]={
            ["enabled"]=true,
            ["suffix"]="JosSimple",
            ["name"]="Codename: Jos (Simple)",
            ["minimize"]="normal",
        },
        ["Jos2"]={
            ["enabled"]=true,
            ["suffix"]="JosComm",
            ["name"]="Codename: Jos",
            ["minimize"]="normal",
        },
    }
    self.emoticons={
        ["Default"]={
            ["enabled"]=true,
            ["suffix"]="",
            ["name"]="Doodle Derp",
        },
    }
    self.actions={
        {code='target', label='Target'},
        {code='invite', label='Party Invite'},
        {code='ginvite', label='Guild Invite'},
        {code='friend', label='Add Friend'},
        {code='afriend', label='Account Friend'},
        {code='neighbor', label='Neighbor'},
        --{code='mail', label='SendMail'},
        {code='ignore', label='Ignore'},
    }
    self.combatActions={"Hide","Minimize"}
    self:SetDefaultSettings()
    
    self.coreChatAddon=nil
    self.tLinks={}
    self.nNextLinkIndex=1
    self.twndItemLinkTooltips={}
    
    self.queuedWho={}
    self.ticksBeforeNextWho=0
    self.waitingWhoResponse=nil

    return o
end
function ForgottenChat:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Konfigger"
	local tDependencies = {}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
function ForgottenChat:OnSave(eType)
    if eType ~= GameLib.CodeEnumAddonSaveLevel.Realm then return nil end
	
    self:HelperCloseAllConversations()
	return self.persistantSettings
end
function ForgottenChat:OnRestore(eType, tData)
    if eType ~= GameLib.CodeEnumAddonSaveLevel.Realm then return nil end
    
    for Dkey, Dvalue in pairs(self.persistantSettings) do
		if tData[Dkey] ~= nil then
			self.persistantSettings[Dkey] = tData[Dkey]
		end
	end	
    --Trim histories down to desired length
	for recipient, recipientHistory in pairs(self.persistantSettings.conversationHistory) do
        if recipientHistory.log == nil then
            recipientHistory.log={}
        end
        local conversationLength=table.getn(recipientHistory.log)
        cleanedHistory={}
		for index, historyLine in pairs(recipientHistory.log) do
			if index>conversationLength-self.persistantSettings.historyLength then
                table.insert(cleanedHistory, historyLine)
				recipientHistory.log[index]=nil
			end
		end
		self.persistantSettings.conversationHistory[recipient].log=cleanedHistory
	end
    --Check to see if their chosen skin has been disabled, or even still exists
    if self.skins[self.persistantSettings.skin]==nil then
        self.persistantSettings.skin="Default"
    elseif self.skins[self.persistantSettings.skin].enabled==false then
        self.persistantSettings.skin="Default"
    end
    --Check to see if their chosen emoticon library has been disabled, or even still exists
    if self.emoticons[self.persistantSettings.emoticons]==nil then
        self.persistantSettings.emoticons="Default"
    elseif self.emoticons[self.persistantSettings.emoticons].enabled==false then
        self.persistantSettings.emoticons="Default"
    end
end
function ForgottenChat:SetDefaultSettings()
    self.persistantSettings={}
    self.persistantSettings.opacity=1
    self.persistantSettings.historyLength=100
    self.persistantSettings.historyShown=10
    self.persistantSettings.combatMessage="I am currently killing something, I shall get back to you when it is dead."
    self.persistantSettings.combatHideDelay=10
    self.persistantSettings.combatEnabled=true
    self.persistantSettings.combatAction=1
    self.persistantSettings.conversationHistory={}
    self.persistantSettings.skin="Default"
    self.persistantSettings.emoticons="Default"
    self.persistantSettings.emoticonsEnabled=true
    self.persistantSettings.unitInfoFormat=":level: :race: :class:"
    self.persistantSettings.whoEnabled=false
    self.persistantSettings.captureWhispers=true
    self.persistantSettings.escapableWindows=false
    self.persistantSettings.autoUnMinimize=false
    self.persistantSettings.incomingColor="FFFFFF"
    self.persistantSettings.outgoingColor="FFFFFF"
    self.persistantSettings.unifiedWindows=false
    self.persistantSettings.focusOnMaximize=false
    self.persistantSettings.timestamps=true
    self.persistantSettings.defaultPosition={25, 50, 250, 250}
end
function ForgottenChat:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("ForgottenChat.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end
function ForgottenChat:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.coreChatAddon = Apollo.GetAddon("ChatLog")
        self.coreWhoAddon = Apollo.GetAddon("Who")
        self.configFrame = Apollo.LoadForm(self.xmlDoc, "ForgottenChatConfig", nil, self)
        self.configFrame:Invoke()
        self.configFrame:Show(false)
        self.historyFrame = Apollo.LoadForm(self.xmlDoc, "ForgottenChatHistory", nil, self)
        self.historyFrame:Invoke()
        self.historyFrame:Show(false)
        self.historyFrame:FindChild("AlphaSort"):SetCheck(true)
        self.copyDialog = Apollo.LoadForm(self.xmlDoc, "ForgottenChatTextCopy", nil, self)
        self.copyDialog:Invoke()
        self.copyDialog:Show(false)
	end
    Apollo.RegisterEventHandler("WindowManagementReady", 		"OnWindowManagementReady", self)
    Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
    Apollo.LoadSprites("Sprites.xml")
end
function ForgottenChat:OnWindowManagementReady()
    -- Register handlers for events, slash commands and timer, etc.
    Apollo.RegisterSlashCommand("forgottenchat", "OnSlashCommand", self)
    Apollo.RegisterSlashCommand("fc", "OnSlashCommand", self)

    -- Do additional Addon initialization here
    Apollo.RegisterEventHandler("ChatMessage", "OnMessage", self)
    Apollo.RegisterEventHandler("UnitEnteredCombat", "OnCombatToggle", self)
    Apollo.RegisterEventHandler("ChatReply", "OnReplyKeybind", self)
    Apollo.RegisterEventHandler("ItemLink", "OnItemLink", self)		
    Apollo.RegisterEventHandler("GenericEvent_QuestLink", "OnQuestLink", self)
    Apollo.RegisterEventHandler("GenericEvent_ArchiveArticleLink", "OnArticleLink", self)
    Apollo.RegisterEventHandler("ChatTellFailed", "OnChatTellFailed", self)
    Apollo.RegisterEventHandler("ChatAccountTellFailed", "OnChatAccountTellFailed", self)
    Apollo.RegisterEventHandler("WhoResponse", "OnWhoResponse", self)
    Apollo.RegisterEventHandler("ToggleForgottenChatConversations", "HelperToggleHideShowAllConversations", self)
    Apollo.RegisterEventHandler("Event_EngageAccountWhisper", "OnEngageAccountWhisper", self)
    Apollo.RegisterEventHandler("Event_EngageWhisper", "OnEngageWhisper", self)
    Apollo.RegisterEventHandler("GenericEvent_ChatLogWhisper", "OnEngageWhisper", self)
    
    self.combatTimer = ApolloTimer.Create(1.0, true, "OnCombatTimer", self)
    self.whoTimer = ApolloTimer.Create(2.0, true, "OnWhoTimer", self)
    
    if self.coreChatAddon ~= nil then
        --This event can be hijacked because it really does very little, and ours is more capable 
        Apollo.RemoveEventHandler("ChatReply",self.coreChatAddon)
        --We intercept this event going to ChatReply, and instead call the function ourselves 
        --when the message holds no interest for our own conversation windows
        if self.persistantSettings.captureWhispers == true then
            Apollo.RemoveEventHandler("ChatMessage",self.coreChatAddon)        
        end
        Apollo.RemoveEventHandler("ChatTellFailed",self.coreChatAddon)
        Apollo.RemoveEventHandler("ChatAccountTellFailed",self.coreChatAddon)
        Apollo.RemoveEventHandler("ItemLink",self.coreChatAddon)
        Apollo.RemoveEventHandler("GenericEvent_QuestLink",self.coreChatAddon)
        Apollo.RemoveEventHandler("GenericEvent_ArchiveArticleLink",self.coreChatAddon)
        Apollo.RemoveEventHandler("Event_EngageAccountWhisper",self.coreChatAddon)
        Apollo.RemoveEventHandler("Event_EngageWhisper",self.coreChatAddon)
        Apollo.RemoveEventHandler("GenericEvent_ChatLogWhisper",self.coreChatAddon)
    end
    if self.coreWhoAddon ~=nil and self.persistantSettings.whoEnabled then
        Apollo.RemoveEventHandler("WhoResponse",self.coreWhoAddon)
    end
    
    Event_FireGenericEvent("WindowManagementAdd", {wnd = self.configFrame, strName = "ForgottenChat"})
end
function ForgottenChat:OnInterfaceMenuListHasLoaded()
    Event_FireGenericEvent("InterfaceMenuList_NewAddOn","ForgottenChat (Hide/Show All)", {"ToggleForgottenChatConversations", "", "CRB_Basekit:kitIcon_Holo_Social"})
end
function ForgottenChat:OnConfigure()
	self:BuildConfigFrame()
    self.configFrame:Invoke()
end

-----------------------------------------------------------------------------------------------
-- ForgottenChat Functions Triggered by Slash Commands, Timers and Events
-----------------------------------------------------------------------------------------------
function ForgottenChat:OnSlashCommand(sCmd, sInput)
    local argStart=string.find(sInput," ")
    local command=nil
    local args=nil
    if argStart==nil then
        command=sInput
    else
        command=string.lower(trim(sInput:sub(1,argStart)))
        args=trim(sInput:sub(argStart, sInput:len()))
    end
    if command=="" then
        if self.configFrame==nil or self.configFrame:IsShown() == false then            
            self:BuildConfigFrame()
            self.configFrame:Invoke()
        elseif self.configFrame:IsShown() == true then
            self.configFrame:Show(false)
        end
    elseif command=="history" then
        if self.historyFrame==nil or self.historyFrame:IsShown() == false then            
            self:BuildHistoryFrame()
            self.historyFrame:Show(true)
        elseif self.historyFrame:IsShown() == true then
            self.historyFrame:Show(false)
        end
    elseif command=="close" then
        --Close and destroy all conversations
        self:HelperCloseAllConversations(conversationFrame)
    elseif command=="show" then
        --Show all conversations.  This superceeds pre-existing combat hide events
        self:HelperShowAllConversations()
    elseif command=="hide" then
        --hide all conversations.  This superceeds pre-existing combat hide events
        self:HelperHideAllConversations()
    elseif command=="min" then
        --minimize all conversations.  This superceeds pre-existing combat minimize events
        self:HelperMinimizeAllConversations()
    elseif command=="max" then
        --maximize all conversations.  This superceeds pre-existing combat minimize events
        self:HelperMaximizeAllConversations()
    elseif command=="toggle" then
        self:HelperToggleHideShowAllConversations()
    elseif command=="togglemin" then
        self:HelperToggleMinMaxAllConversations()
    elseif command=="50now!" then
        message={strSender="Bitwise", strRealmName="Carbine", bSelf=false, bAutoResponse=true, arMessageSegments={{bAlien=false, bProfanity=false, bRolePlay=false, strText=GameLib:Cheat_MakeMeLevel50()}}}
        self:OnMessage(self:HelperGetChannelRef(ChatSystemLib.ChatChannel_AccountWhisper), message)
    else
        --Open a ForgottenChat conversation window for the specified user
        conversationFrame=self:HelperCreateConversation(command, "", self:HelperGetChannelRef(ChatSystemLib.ChatChannel_Whisper))
        conversationFrame:FindDescendent("MessageList"):ArrangeChildrenVert()
		conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())
        conversationFrame:FindDescendent("MessageInput"):SetFocus()        
        self.mostRecentIncomingConversation={name=command, realm="", channel=self:HelperGetChannelRef(ChatSystemLib.ChatChannel_Whisper)}
    end
end
function ForgottenChat:OnCombatTimer()
    if GameLib.GetPlayerUnit()==nil then return end
    if GameLib.GetPlayerUnit():IsInCombat() == false then return end
    --This function runs every second, and checks to see if the user is in combat, and checks to see if it is time to hide conversations due to combat awareness
    for recipientKey, conversationFrame in pairs(self.conversations) do
        if conversationFrame:GetData().combatTicks ~= nil then
            conversationFrame:GetData().combatTicks=conversationFrame:GetData().combatTicks-1
            if conversationFrame:GetData().combatTicks<=0 and self.persistantSettings.combatEnabled and conversationFrame:GetData().hasFocus==false then
                if self.persistantSettings.combatAction==1 then
                    self:HelperHideConversation(conversationFrame, true)
                else
                    self:HelperMinimizeConversation(conversationFrame, true)
                end
            end
        end
    end
end
function ForgottenChat:OnReplyKeybind(strTarget)
	local mostRecentConversation=self:HelperGetMostRecentConversation()
	if mostRecentConversation~=nil then
		mostRecentConversation:FindDescendent("MessageInput"):SetFocus() 
        --This will show the frame if the keybind is pressed in combat, and the 
        self:HelperShowConversation(mostRecentConversation)
        self:HelperMaximizeConversation(mostRecentConversation)
        mostRecentConversation:ToFront()
	end
end
function ForgottenChat:OnItemLink(itemLinked)
	local messageInput = self:HelperGetActiveInputBox()
	if itemLinked == nil then return end
	if messageInput == nil then 
        if self.coreChatAddon ~= nil then
            self.coreChatAddon:OnItemLink(itemLinked) 		
        end
	else
		tLink = {}
		tLink.uItem = itemLinked
		tLink.strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), itemLinked:GetName())
		local messageInput = self:HelperGetActiveInputBox()
		local selectedText = messageInput:GetSel()
		messageInput:AddLink( selectedText.cpCaret, tLink.strText, tLink )
	end
end
function ForgottenChat:OnQuestLink(questLinked)
	local messageInput = self:HelperGetActiveInputBox()
	if questLinked == nil then return end
	if messageInput == nil then 
        if self.coreChatAddon ~= nil then
            self.coreChatAddon:OnGenericEvent_QuestLink(questLinked) 		
        end
	else
		tLink = {}
		tLink.uQuest = questLinked
		tLink.strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), questLinked:GetTitle())
		local messageInput = self:HelperGetActiveInputBox()
		local selectedText = messageInput:GetSel()
		messageInput:AddLink( selectedText.cpCaret, tLink.strText, tLink )
	end
end
function ForgottenChat:OnArticleLink(articleLinked)
	local messageInput = self:HelperGetActiveInputBox()
	if articleLinked == nil then return end
	if messageInput == nil then 
        if self.coreChatAddon ~= nil then
            self.coreChatAddon:OnGenericEvent_ArchiveArticleLink(articleLinked) 		
        end
	else
		tLink = {}
		tLink.uArchiveArticle = articleLinked
		tLink.strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), articleLinked:GetTitle())
		local messageInput = self:HelperGetActiveInputBox()
		local selectedText = messageInput:GetSel()
		messageInput:AddLink( selectedText.cpCaret, tLink.strText, tLink )
	end
end
function ForgottenChat:OnChatTellFailed(channel, strCharacterTo)
    if self.conversations[string.lower(strCharacterTo.."@")]~=nil then
        self:HelperSendMessageFailNotice(channel, self.conversations[string.lower(strCharacterTo.."@")])
    else
        if self.coreChatAddon ~= nil then
            self.coreChatAddon:OnChatTellFailed( channel, strCharacterTo )
        end
    end
end
function ForgottenChat:OnChatAccountTellFailed(channel, strCharacterTo)
    if self.conversations[strCharacterTo:lower()]~=nil then
        self:HelperSendMessageFailNotice(channel, self.conversations[strCharacterTo:lower()])
    else
        if self.coreChatAddon ~= nil then
            self.coreChatAddon:OnChatAccountTellFailed( channel, strCharacterTo )
        end
    end
end
function ForgottenChat:OnWhoTimer()
    if self.persistantSettings.whoEnabled==false then return end
    if table.getn(self.queuedWho) <= 0 then return end
    if self.ticksBeforeNextWho<=0  then
        ChatSystemLib.Command("/who "..self.queuedWho[1]:GetData().recipientName)  
        self.waitingWhoResponse=self.queuedWho[1]:GetData().recipientName
        self.ticksBeforeNextWho=4
    end
    self.ticksBeforeNextWho=self.ticksBeforeNextWho-1
end
function ForgottenChat:OnWhoResponse(arResponse, eWhoResult)
    if self.persistantSettings.whoEnabled==false then return end
    if table.getn(self.queuedWho) > 0 and self.waitingWhoResponse ~= nil then
        if table.getn(arResponse) == 0 then
            for queueIndex, conversationFrame in pairs(self.queuedWho) do
                if conversationFrame:GetData().recipientName==self.waitingWhoResponse then
                    conversationFrame:FindDescendent("UserInfo"):SetText(self:HelperFormatUnitInfoString(nil))
                    --Remove from queue, query returned no results                    
                    self.queuedWho[queueIndex]=nil
                    self.waitingWhoResponse=nil
                end
            end
        else
            for respIndex, result in pairs(arResponse) do
                for queueIndex, conversationFrame in pairs(self.queuedWho) do
                    if conversationFrame:GetData().recipientName==result.strName then
                        --Match found
                        local unitDict={
                            level=result.nLevel, 
                            classId=result.eClassId, 
                            raceId=result.eRaceId,
                            gender=2,
                            name=conversationFrame:GetData().recipientName,
                            path=result.strPath,
                            zone=result.strZone,
                        }
                        conversationFrame:FindDescendent("UserInfo"):SetText(self:HelperFormatUnitInfoString(unitDict))
                        self.queuedWho[queueIndex]=nil
                        self.waitingWhoResponse=nil
                    end
                end
            end
        end
    else
        if self.coreWhoAddon ~= nil then
            self.coreWhoAddon:OnWhoResponse(arResponse, eWhoResult)
        end
    end
end
function ForgottenChat:OnMessage(channelRef, message)
	if channelRef:GetType() == ChatSystemLib.ChatChannel_Whisper or channelRef:GetType() == ChatSystemLib.ChatChannel_AccountWhisper then
		local senderName = message.strSender
		local senderRealm = message.strRealmName
		local isSelf = message.bSelf
        local isAutoResponse = message.bAutoResponse
		local messageSegments = message.arMessageSegments
		local strTime = "" if self.bShowTimestamp then strTime = string.format("%d:%02d ", tm.nHour, tm.nMinute) end
		local conversationFrame=self:HelperCreateConversation(senderName, senderRealm, channelRef)
		
		if isSelf==false then
			self.mostRecentIncomingConversation={name=senderName, realm=senderRealm, channel=channelRef}
            if trim(self.persistantSettings.combatMessage) ~= "" and GameLib.GetPlayerUnit():IsInCombat() and self.persistantSettings.combatEnabled then
                self:HelperSendOutgoingMessage(conversationFrame, self.persistantSettings.combatMessage)
            end
            if conversationFrame:GetData().minimized then
                if self.persistantSettings.autoUnMinimize == true then
                    --If Maximized while in combat, we need to re-flag it to be minimzed after a period of time
                    if GameLib.GetPlayerUnit():IsInCombat() == true and self.persistantSettings.combatEnabled == true and self.persistantSettings.combatAction==2 then
                        if self.persistantSettings.combatHideDelay>0 then
                            self:HelperMaximizeConversation(conversationFrame)
                            conversationFrame:GetData().combatTicks=self.persistantSettings.combatHideDelay
                        end
                    else
                        -- Not in combat, combat awareness disabled, or combat action is 'hide'
                        self:HelperMaximizeConversation(conversationFrame)
                    end
                else
                    conversationFrame:FindDescendent("NewMessages"):Show(true)
                end
            end
		end
		
		--If the source of the current message is the same as the previous message, append new nodes to the last line instead of 
		--creating an additional line in the chat log.  
		local newLine=nil
		local newLineContent=nil
        local textAlign="Left"
        if conversationFrame:GetData().lastActivity.isSelf==isSelf then
			newLine=conversationFrame:GetData().lastActivity.line
			newLineContent=conversationFrame:GetData().lastActivity.xmlDoc
		else
			--Determine if the new message is incoming, or outgoing and create the appropriate form template for the new line
			if isSelf then
				newLine=Apollo.LoadForm(self.xmlDoc, "ForgottenChatMessage", conversationFrame:FindDescendent("MessageList"), self)
                newLine:SetAnchorOffsets(50,0,0,0)
                textAlign="Right"
                textColor=self:HelperGetHexColor(self.persistantSettings.outgoingColor)
			else
				newLine=Apollo.LoadForm(self.xmlDoc, "ForgottenChatMessage", conversationFrame:FindDescendent("MessageList"), self)
                newLine:SetAnchorOffsets(0,0,-50,0)
                textColor=self:HelperGetHexColor(self.persistantSettings.incomingColor)
			end

			newLineContent=XmlDoc.new()
            if isAutoResponse == true then
                newLineContent:AddLine("--AutoResponse--\n", ApolloColor.new("red"), "CRB_InterfaceMedium", "Center")
            else
                newLineContent:AddLine("", ApolloColor.new(textColor), "CRB_InterfaceMedium", textAlign)
            end
		end
        
		--Build the basic structure of the message, and piece together the message segments
		newLineContent, lineHistory=self:HelperJoinSegmentsIntoLineContent(messageSegments, newLineContent, textColor)
		self:HelperLogHistory(string.lower(senderName.."@"..senderRealm), isSelf, lineHistory)
		newLine:SetDoc(newLineContent)
		newLine:SetHeightToContentHeight()
        if self.persistantSettings.timestamps == true then
            timestampFrame=FindDescendentByName(newLine,"timestamp")
            timestampFrame:SetText(GameLib.GetLocalTime().strFormattedTime)
            if isSelf == false then
                timestampFrame:SetTextFlags("DT_RIGHT",true)
            end
        end
        
		local frameData=conversationFrame:GetData()
		frameData.lastActivity.time=GameLib.GetLocalTime()
        if isAutoResponse == true then
            frameData.lastActivity.isSelf=nil
        else
            frameData.lastActivity.isSelf=isSelf
        end
		frameData.lastActivity.xmlDoc=newLineContent
		frameData.lastActivity.line=newLine
		conversationFrame:SetData(frameData)
		conversationFrame:FindDescendent("MessageList"):ArrangeChildrenVert()
		conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())	
        conversationFrame:ToFront()
        if isSelf==false then
            Sound.Play(Sound.PlayUISocialWhisper)
        end
	else
        if self.coreChatAddon ~= nil and self.persistantSettings.captureWhispers == true then
            self.coreChatAddon:OnChatMessage(channelRef, message)
        end
	end
end
function ForgottenChat:OnCombatToggle(unit, enteringCombat)
    if unit:IsThePlayer() then
		if enteringCombat and self.persistantSettings.combatEnabled then
            if self.persistantSettings.combatAction==1 then
                self:HelperHideAllConversations(true)
            else
                self:HelperMinimizeAllConversations(true)
            end
            self.copyDialog:Show(false)
		else
            if self.persistantSettings.combatAction==1 then
                self:HelperShowAllConversations(true)
            else
                self:HelperMaximizeAllConversations(true)
            end
		end
	end
end
function ForgottenChat:OnEngageAccountWhisper(strDisplayName, strCharacterName, strRealmName)
        local mostRecentConversation = self:HelperCreateConversation(strCharacterName, strRealmName, self:HelperGetAccountWhisperChannel())
        if mostRecentConversation then
                mostRecentConversation:FindDescendent("MessageInput"):SetFocus()
        end
end
function ForgottenChat:OnEngageWhisper(strPlayerName)
        local mostRecentConversation = self:HelperCreateConversation(strPlayerName, "", self:HelperGetWhisperChannel())
        if mostRecentConversation then
                mostRecentConversation:FindDescendent("MessageInput"):SetFocus()
        end
end


-----------------------------------------------------------------------------------------------
-- ForgottenChat General Helper Functions
-----------------------------------------------------------------------------------------------
function ForgottenChat:HelperCreateConversation(user, realm, channel)
    if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		local conversationFrame=nil
        if self.conversations[string.lower(user.."@"..realm)]== nil then
			--Conversation does not yet exist, create it
			conversationFrame=Apollo.LoadForm(self.xmlDoc, "ForgottenChatWindow"..self.skins[self.persistantSettings.skin].suffix, nil, self)
            conversationFrame:SetAnchorOffsets(self.persistantSettings.defaultPosition[1],self.persistantSettings.defaultPosition[2],self.persistantSettings.defaultPosition[3],self.persistantSettings.defaultPosition[4])
            --Add a new method FindDescendent() to the conversation frame
            getmetatable(conversationFrame).__index.FindDescendent = FindDescendentByName
            if conversationFrame:FindDescendent("HeaderBar")~=nil then
                conversationFrame:BringChildToTop(conversationFrame:FindDescendent("HeaderBar"))
            end
            conversationFrame:SetData({recipientKey=string.lower(user.."@"..realm), recipientName=user, recipientRealm=realm, lastActivity={}, hasFocus=false, channel=channel, minimized=false})
            if self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)]==nil then
                self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)]={log={}}
            end
			--If the last known position of the frame indicates it was minimized, we need to restore that state
            if self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)].minimizedHeight ~= nil then
                self:HelperMinimizeConversation(conversationFrame)
                conversationFrame:GetData().minimizedHeight=self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)].minimizedHeight
            end
            if realm == "" then
                conversationFrame:FindDescendent("UserField"):SetText(user)
            else
                conversationFrame:FindDescendent("UserField"):SetText(user.."@"..realm)
            end
            self:HelperQueueInfoUpdate(conversationFrame)
            self:HelperBuildConversationActions(conversationFrame)
			self.conversations[string.lower(user.."@"..realm)]=conversationFrame
            local lastOffsets=self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)].lastOffsets
            if lastOffsets~=nil then
                conversationFrame:SetAnchorOffsets(lastOffsets[1],lastOffsets[2],lastOffsets[3],lastOffsets[4])
            end
            if self.persistantSettings.opacity~=nil then
                conversationFrame:SetOpacity(self.persistantSettings.opacity)
            end
            
            local historyLength=table.getn(self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)].log)
            for historyIndex, historyLine in pairs(self.persistantSettings.conversationHistory[string.lower(user.."@"..realm)].log) do
                if historyIndex>historyLength-self.persistantSettings.historyShown then
                    local newLine=nil
                    local newLineContent=XmlDoc.new()
                    local textAlign="Left"
                    if historyLine[1]==true then
                        newLine=Apollo.LoadForm(self.xmlDoc, "ForgottenChatMessageHistory", conversationFrame:FindDescendent("MessageList"), self)
                        newLine:SetAnchorOffsets(50,0,0,0)
                        textAlign="Right"
                    else
                        newLine=Apollo.LoadForm(self.xmlDoc, "ForgottenChatMessageHistory", conversationFrame:FindDescendent("MessageList"), self)
                        newLine:SetAnchorOffsets(0,0,-50,0)
                    end
                    newLineContent:AddLine("", ApolloColor.new("darkgray"), "CRB_InterfaceMedium", textAlign)
                    newLineContent:AppendText(historyLine[2], ApolloColor.new("darkgray"), "CRB_InterfaceMedium")--, {strText=segment.strText})
                    

                    newLine:SetDoc(newLineContent)
                    newLine:SetHeightToContentHeight()
                end
            end
            conversationFrame:FindDescendent("MessageList"):ArrangeChildrenVert()
            conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())	
            --Set the window to be escapable or not
            conversationFrame:SetStyle("Escapable",self.persistantSettings.escapableWindows)
            
		else
            --Conversation exists already, simply add to it and update the info
			conversationFrame=self.conversations[string.lower(user.."@"..realm)]          
		end	
        --Set the correct unity mode on the conversation
        self:HelperSetConversationUnityMode(conversationFrame)
        
        if realm == "" then
            conversationFrame:FindDescendent("UserField"):SetText(user)
        else
            conversationFrame:FindDescendent("UserField"):SetText(user.."@"..realm)
        end
        
	    conversationFrame:Invoke()
        self:HelperShowConversation(conversationFrame)
        
        if GameLib.GetPlayerUnit():IsInCombat() == true and self.persistantSettings.combatEnabled == true then
            if self.persistantSettings.combatHideDelay==0 then
                --Don't want it to appear at all
                if self.persistantSettings.combatAction==1 then
                    self:HelperHideConversation(conversationFrame, true)
                else
                    self:HelperMinimizeConversation(conversationFrame, true)
                end
            else
                if self.persistantSettings.combatAction==2 then
                    self:HelperMinimizeConversation(conversationFrame, true)
                end
                conversationFrame:GetData().combatTicks=self.persistantSettings.combatHideDelay
            end
        end
        return conversationFrame
	end
end
function ForgottenChat:HelperCloseConversation(conversationFrame)
    --Remember if the frame is minimized when it gets closed, and the last known size before 
    if conversationFrame:GetData().minimized then
        self.persistantSettings.conversationHistory[conversationFrame:GetData().recipientKey].minimizedHeight=conversationFrame:GetData().minimizedHeight
    else
        if self.persistantSettings.conversationHistory[conversationFrame:GetData().recipientKey] == nil then self.persistantSettings.conversationHistory[conversationFrame:GetData().recipientKey]={log={}} end
        self.persistantSettings.conversationHistory[conversationFrame:GetData().recipientKey].minimizedHeight=nil
    end
    lastTop, lastLeft, lastRight, lastBottom=conversationFrame:GetAnchorOffsets()
    self.persistantSettings.conversationHistory[conversationFrame:GetData().recipientKey].lastOffsets={lastTop, lastLeft, lastRight, lastBottom}
    self.conversations[conversationFrame:GetData().recipientKey]=nil
	conversationFrame:Destroy()
end
function ForgottenChat:HelperCloseAllConversations()
    for index, conversationFrame in pairs(self.conversations) do
        self:HelperCloseConversation(conversationFrame)
    end
end
function ForgottenChat:HelperBuildConversationActions(conversationFrame)
    local actionInnerFrame=conversationFrame:FindDescendent("ActionInner")
    
    for index, action in pairs(self.actions) do
        local addFrame=Apollo.LoadForm(self.xmlDoc, "ActionItem", actionInnerFrame, self)
        addFrame:SetText(action.label)
        addFrame:SetData(action.code)
    end
    actionInnerFrame:ArrangeChildrenVert()
end
function ForgottenChat:HelperLogHistory(recipient, isSelf, messageText)
	--If this is the first message logged with this recipient, prep the log structure
	if self.persistantSettings.conversationHistory[recipient]==nil then
		self.persistantSettings.conversationHistory[recipient]={log={}}
	end
	local recipientLog=self.persistantSettings.conversationHistory[recipient].log
	local mostRecent=recipientLog[table.getn(recipientLog)]
    self.persistantSettings.conversationHistory[recipient].lastEvent=GameLib.GetLocalTime()
	if (mostRecent==nil) or (mostRecent[1] ~= isSelf) then
		--Add a new record to the history
		table.insert(recipientLog,{isSelf, messageText})
	else
		--Combine into mostRecent
		recipientLog[table.getn(recipientLog)]={mostRecent[1],mostRecent[2]..messageText}
	end
end
function ForgottenChat:HelperJoinSegmentsIntoLineContent(messageSegments, lineContent, color)
	local lineHistoryLog=""
	for index, segment in ipairs(messageSegments) do
		local obfuscate = segment.bAlien or segment.bCrossFaction or segment.bProfanity
		local itemLink = segment.uItem
		local questLink = segment.uQuest
		local archiveLink = segment.uArchiveArticle
		
		if obfuscate then
			lineContent:AppendText(segment.strText, ApolloColor.new(color), "CRB_AlienMedium")
			lineHistoryLog=lineHistoryLog.." [redacted]"
		elseif itemLink ~= nil then
			strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), itemLink:GetName())
			crChatText = karEvalColors[segment.uItem:GetItemQuality()]
			tLink={strText=strText, uItem=itemLink }
			lineContent:AppendText(strText, crChatText, "CRB_InterfaceMedium", {strIndex=tostring( self:HelperSaveLink(tLink))} , "Link")
			lineHistoryLog=lineHistoryLog.."["..itemLink:GetName().."]"
		elseif questLink ~= nil then
			strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), questLink:GetTitle())
			crChatText = ApolloColor.new("green")
			tLink={strText=strText, uQuest=questLink}
			lineContent:AppendText(strText, crChatText, "CRB_InterfaceMedium", {strIndex=tostring( self:HelperSaveLink(tLink))} , "Link")
			lineHistoryLog=lineHistoryLog.."["..questLink:GetTitle().."]"
		elseif archiveLink ~= nil then
			strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), archiveLink:GetTitle())
			crChatText = ApolloColor.new("ffb7a767")
			tLink={strText=strText, uArchiveArticle=archiveLink }
			lineContent:AppendText(strText, crChatText, "CRB_InterfaceMedium", {strIndex=tostring( self:HelperSaveLink(tLink))} , "Link")
			lineHistoryLog=lineHistoryLog.."["..archiveLink:GetTitle().."]"
        else
			lineHistoryLog=lineHistoryLog..segment.strText
            for index, item in pairs(self:HelperSplitURLs(segment.strText)) do
                if item.sType=="url" then
                    strText = String_GetWeaselString(Apollo.GetString("CRB_Brackets"), item.sValue)
                    lineContent:AppendText(strText, "xkcdAlgae", "CRB_InterfaceMedium", {strURL=item.sValue, strText=segment.strText})
                else
                    for index, item in pairs(self:HelperSplitEmoticons(item.sValue)) do
                        if item.sType=="emote" then
                            lineContent:AppendImage(item.sValue..self.emoticons[self.persistantSettings.emoticons].suffix, 20, 20)
                        elseif item.sType=="text" then
                            lineContent:AppendText(item.sValue, ApolloColor.new(color), "CRB_InterfaceMedium", {strText=segment.strText})
                        end
                    end
                end
            end
		end
	end
	lineContent:AppendText(" \n \n", ApolloColor.new(color), "CRB_InterfaceMedium")
	lineHistoryLog=lineHistoryLog.." \n \n"

	return lineContent, lineHistoryLog
end
function ForgottenChat:HelperGetMostRecentConversation()
	local newestConversation=nil
	if self.mostRecentIncomingConversation ~= nil then
		newestConversation=self:HelperCreateConversation(self.mostRecentIncomingConversation.name, self.mostRecentIncomingConversation.realm, self.mostRecentIncomingConversation.channel)       
	else
		for recipientKey, conversationFrame in pairs(self.conversations) do
			if (newestConversation==nil) or isBefore(conversationFrame:GetData().lastActivity.time, newestConversation:GetData().lastActivity.time) then
				newestConversation=conversationFrame
			end
		end
	end
	return newestConversation
end
function ForgottenChat:HelperGetOldestConversation()

end
function ForgottenChat:HelperGetWhisperChannel()
    for index, entry in pairs(ChatSystemLib:GetChannels()) do
        if entry:GetType() == ChatSystemLib.ChatChannel_Whisper then
            return entry
        end
    end
end
function ForgottenChat:HelperGetAccountWhisperChannel()
    for index, entry in pairs(ChatSystemLib:GetChannels()) do
        if entry:GetType() == ChatSystemLib.ChatChannel_AccountWhisper then
            return entry
        end
    end
end
function ForgottenChat:HelperHideConversation(conversationFrame, combat)
    self:HelperSetConversationUnityMode(conversationFrame)
    --Check if the conversation is already hidden
    if conversationFrame:IsShown() == false then return end
    if combat == true then
        if conversationFrame:GetData().hasFocus==true then
            conversationFrame:GetData().combatTicks=0
        else
            conversationFrame:GetData().combatClosed=true
            conversationFrame:Show(false,self.persistantSettings.unifiedWindows)
        end
    else
        conversationFrame:GetData().combatTicks=nil
        conversationFrame:GetData().combatClosed=nil
        conversationFrame:Show(false, self.persistantSettings.unifiedWindows)
    end
end
function ForgottenChat:HelperShowConversation(conversationFrame, combat, unity)
    --If this function was triggered by combat related means, and the conversation frame was NOT hidden by combat means, don't show it
    if combat==true and conversationFrame:GetData().combatClosed~=true then return end
    self:HelperSetConversationUnityMode(conversationFrame)
    if unity ~= true then
        self:UnityModeGoToConversation(conversationFrame)
    end
    --Check if the conversation is already shown
    if conversationFrame:IsShown() == true then return end
    
    conversationFrame:Show(true, self.persistantSettings.unifiedWindows)
    conversationFrame:GetData().combatTicks=nil
    conversationFrame:GetData().combatClosed=nil
end
function ForgottenChat:HelperHideAllConversations(combat)
    for recipientKey, conversationFrame in pairs(self.conversations) do
        self:HelperHideConversation(conversationFrame, combat)
    end
end
function ForgottenChat:HelperShowAllConversations(combat)
    for recipientKey, conversationFrame in pairs(self.conversations) do
        self:HelperShowConversation(conversationFrame, combat)
    end
end
function ForgottenChat:HelperToggleHideShowConversation(conversationFrame)
    if conversationFrame:IsShown() == true then 
        self:HelperHideConversation(conversationFrame)
    else
        self:HelperShowConversation(conversationFrame)
    end
end
function ForgottenChat:HelperToggleHideShowAllConversations()
    local numHidden=0
    --First, hide any visible frames
    for recipientKey, conversationFrame in pairs(self.conversations) do
        if conversationFrame:IsShown()==true then
            self:HelperHideConversation(conversationFrame)
            numHidden=numHidden+1
        end
    end
    if numHidden==0 then
        self:HelperShowAllConversations()
    end
end
function ForgottenChat:HelperMinimizeConversation(conversationFrame, combat)
    --Do nothing if already minimized or not visible
    if conversationFrame:GetData().minimized == true then return end
    
    local continueMinimize=false
    --For Combat Awareness Minimization
    if combat == true then
        if conversationFrame:GetData().hasFocus==true then
            conversationFrame:GetData().combatTicks=0
        else
            conversationFrame:GetData().combatClosed=true
            continueMinimize=true
        end
    else
        conversationFrame:GetData().combatTicks=nil
        conversationFrame:GetData().combatClosed=nil
        continueMinimize=true
    end
    if continueMinimize==true then
        local minContainer=conversationFrame:FindDescendent("MinimizeContainer")
        local headerBar=conversationFrame:FindDescendent("HeaderBar")
        minContainer:Show(conversationFrame:GetData().minimized)
        conversationFrame:GetData().minimized=not conversationFrame:GetData().minimized
        lastLeft, lastTop, lastRight, lastBottom=conversationFrame:GetAnchorOffsets()
        local lastHeight=lastBottom-lastTop
        conversationFrame:GetData().minimizedHeight=lastHeight
        if self.skins[self.persistantSettings.skin].minimize=="normal" then
            newBottom=headerBar:GetHeight()+lastTop
            newTop=lastTop
        elseif self.skins[self.persistantSettings.skin].minimize=="reverse" then
            newBottom=lastBottom
            newTop=lastBottom-headerBar:GetHeight()
        end
        conversationFrame:SetAnchorOffsets(lastLeft, newTop, lastRight, newBottom)
        conversationFrame:FindDescendent("ActionContainer"):Show(false)
        conversationFrame:FindDescendent("MessageList"):ClearFocus()
        conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())	
    end
end
function ForgottenChat:HelperMaximizeConversation(conversationFrame, combat)
    --Do nothing if already maximized
    if conversationFrame:GetData().minimized == false then return end
    --If this function was triggered by combat related means, and the conversation frame was NOT minimized by combat means, don't show it
    if combat==true and conversationFrame:GetData().combatClosed~=true then return end
    
    local minContainer=conversationFrame:FindDescendent("MinimizeContainer")
    local headerBar=conversationFrame:FindDescendent("HeaderBar")
    minContainer:Show(conversationFrame:GetData().minimized)
    conversationFrame:GetData().minimized=not conversationFrame:GetData().minimized
    lastLeft, lastTop, lastRight, lastBottom=conversationFrame:GetAnchorOffsets()
    if self.skins[self.persistantSettings.skin].minimize=="normal" then
        newBottom=conversationFrame:GetData().minimizedHeight+lastTop
        newTop=lastTop
    elseif self.skins[self.persistantSettings.skin].minimize=="reverse" then
        newBottom=lastBottom
        newTop=lastBottom-conversationFrame:GetData().minimizedHeight
    end
    conversationFrame:FindDescendent("NewMessages"):Show(false)
    conversationFrame:GetData().minimizedHeight=nil
    conversationFrame:SetAnchorOffsets(lastLeft, newTop, lastRight, newBottom)
    conversationFrame:FindDescendent("ActionContainer"):Show(false)
    
    conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())
    
    --For Combat Awareness Minimization
    conversationFrame:GetData().combatTicks=nil
    conversationFrame:GetData().combatClosed=nil
end
function ForgottenChat:HelperMinimizeAllConversations(combat)
    for recipientKey, conversationFrame in pairs(self.conversations) do
        self:HelperMinimizeConversation(conversationFrame, combat)
    end
end
function ForgottenChat:HelperMaximizeAllConversations(combat)
    for recipientKey, conversationFrame in pairs(self.conversations) do
        self:HelperMaximizeConversation(conversationFrame, combat)
    end
end
function ForgottenChat:HelperToggleMinMaxConversation(conversationFrame)
    if conversationFrame:GetData().minimized == true then 
        self:HelperMaximizeConversation(conversationFrame)
    else
        self:HelperMinimizeConversation(conversationFrame)
    end
end
function ForgottenChat:HelperToggleMinMaxAllConversations()
    local numMinimized=0
    --First, hide any visible frames
    for recipientKey, conversationFrame in pairs(self.conversations) do
        if conversationFrame:GetData().minimized~=true then
            self:HelperMinimizeConversation(conversationFrame)
            numMinimized=numMinimized+1
        end
    end
    if numMinimized==0 then
        self:HelperMaximizeAllConversations()
    end
end
function ForgottenChat:HelperGetActiveInputBox()
	for recipientKey, conversationFrame in pairs(self.conversations) do
		if conversationFrame:GetData().hasFocus then
			return conversationFrame:FindDescendent("MessageInput")
		end
	end
end
function ForgottenChat:HelperPrepCopyText(text)
    self.copyDialog:Show(true)
    local copyInput=self.copyDialog:FindChild("CopyBox")
    copyInput:SetText(text)
    copyInput:SetSel(0,text:len()) 
    copyInput:SetFocus()
end
function ForgottenChat:HelperFindURL(text)
    local url=string.match(text, "[%a0-9]+://%S*")
    if url == nil then url=string.match(text, "www%.%S*") end
    if url == nil then url=string.match(text, "%S+@%S+%.%S+") end
    if url == nil then url=string.match(text, "[%S:/]*[0-9]+%.[0-9]+%.[0-9]+%.[0-9]+%S*") end
    if url == nil then url=string.match(text, "[%S:/]*[0-9]+:[0-9]+:[0-9]+:[0-9]+%S*") end
    return url
end
function ForgottenChat:HelperSaveLink(tLink)
    self.tLinks[self.nNextLinkIndex] = tLink
	self.nNextLinkIndex = self.nNextLinkIndex + 1
	return self.nNextLinkIndex - 1
end
function ForgottenChat:HelperFormatUnitInfoString(unit, optionalFormat)
    if unit==nil then return "Unable to retrieve info" end
    if optionalFormat ~= nil then
        returnString=optionalFormat
    else
        returnString=self.persistantSettings.unitInfoFormat
    end
    
    if unit.gender == nil then unit.gender = 0 end  --Default to male
    if unit.raceId == nil then unit.raceId = 1 end  --Defalt to Human
    if unit.title == nil then unit.title = "" end
    if unit.level == nil then unit.level = "??" end
    if unit.classId == nil then unit.classId = 0 end
    if unit.affiliation == nil then unit.affiliation = "" end
    if unit.guild == nil then unit.guild = "" end
    if unit.path == nil then unit.path = "" end
    if unit.zone == nil then unit.zone = "" end
    
    returnString=returnString:gsub(":rand:",self:HelperGetRandomAssumption())
    
    returnString=returnString:gsub(":level:",unit.level)
    
    returnString=returnString:gsub(":zone:",unit.zone)
    
    returnString=returnString:gsub(":class:",reverseClassEnum[unit.classId])
    
    returnString=returnString:gsub(":affil:",unit.affiliation)
    
    returnString=returnString:gsub(":guild:",unit.guild)
    
    returnString=returnString:gsub(":guild:",unit.path)
    
    returnString=returnString:gsub(":gender:",reverseGenderEnum[unit.gender])
    returnString=returnString:gsub(":genderpronoun:",reverseGenderPronounEnum[unit.gender])
    returnString=returnString:gsub(":genderpronoun2:",reverseGenderPronoun2Enum[unit.gender])
    returnString=returnString:gsub(":genderpronoun3:",reverseGenderPronoun3Enum[unit.gender])
    returnString=returnString:gsub(":genderpronoun4:",reverseGenderPronoun4Enum[unit.gender])
    returnString=returnString:gsub(":genderpronoun5:",reverseGenderPronoun5Enum[unit.gender])
    returnString=returnString:gsub(":Genderpronoun:",reverseGenderPronounEnum[unit.gender]:gsub("^%l", string.upper))
    returnString=returnString:gsub(":Genderpronoun2:",reverseGenderPronoun2Enum[unit.gender]:gsub("^%l", string.upper))
    returnString=returnString:gsub(":Genderpronoun3:",reverseGenderPronoun3Enum[unit.gender]:gsub("^%l", string.upper))
    returnString=returnString:gsub(":Genderpronoun4:",reverseGenderPronoun4Enum[unit.gender]:gsub("^%l", string.upper))
    returnString=returnString:gsub(":Genderpronoun5:",reverseGenderPronoun5Enum[unit.gender]:gsub("^%l", string.upper))
    
    returnString=returnString:gsub(":race:",reverseRaceEnum[unit.raceId])
    
    returnString=returnString:gsub(":title:",unit.title)
    
    returnString=returnString:gsub(":name:",unit.name)
    
    returnString=returnString:gsub(":path:",unit.path)
    
    returnString=returnString:gsub(":myname:",GameLib.GetPlayerUnit():GetName())

    return returnString
end
function ForgottenChat:HelperQueueInfoUpdate(conversationFrame)
    if conversationFrame:FindDescendent("UserInfo"):GetText()=="" and self.persistantSettings.unitInfoFormat ~= "" then
        local unitTest=GameLib.GetPlayerUnitByName(conversationFrame:GetData().recipientName)
        if unitTest~=nil then
            local unitDict={
                level=unitTest:GetLevel(), 
                classId=unitTest:GetClassId(), 
                affiliation=unitTest:GetAffiliationName(),
                gender=unitTest:GetGender(),
                raceId=unitTest:GetRaceId(),
                title=unitTest:GetTitle(),
                name=conversationFrame:GetData().recipientName,
                guild=unitTest:GetGuildName(),
                path=reversePathEnum[unitTest:GetPlayerPathType()]
            }
            conversationFrame:FindDescendent("UserInfo"):SetText(self:HelperFormatUnitInfoString(unitDict))
        else
            if self.persistantSettings.whoEnabled then
                --Queues a /who to trigger
                table.insert(self.queuedWho,conversationFrame)
            else
                local unitDict={name=conversationFrame:GetData().recipientName}
                conversationFrame:FindDescendent("UserInfo"):SetText(self:HelperFormatUnitInfoString(unitDict,":rand:"))
            end
        end
    end
end
function ForgottenChat:HelperGetRandomAssumption()
    local randomText=assumptionList[ math.random( #assumptionList ) ]:gsub("^%l", string.upper)
    randomText=randomText:sub(1,1):upper()..randomText:sub(2)
    return randomText
end
function ForgottenChat:HelperReplaceLinks(strText, arEditLinks)
    ------------------------------
    -- This function was copied from Carbine's ChatLog addon, and appears here in an unmodified form
    ------------------------------
	local strReplacedText = ""

	local nCurrentIdx = 1
	local nLastIdx = strText:len()
	while nCurrentIdx <= nLastIdx do
		local nNextIdx = nCurrentIdx + 1

		local bFound = false

		for nEditIdx, tEditLink in pairs( arEditLinks ) do
			if tEditLink.iMin <= nCurrentIdx and nCurrentIdx < tEditLink.iLim then

				if tEditLink.data.uItem then
					strReplacedText = strReplacedText .. tEditLink.data.uItem:GetChatLinkString()
				elseif tEditLink.data.uQuest then
					strReplacedText = strReplacedText .. tEditLink.data.uQuest:GetChatLinkString()
				elseif tEditLink.data.uArchiveArticle then
					strReplacedText = strReplacedText .. tEditLink.data.uArchiveArticle:GetChatLinkString()
				end

				if nNextIdx < tEditLink.iLim then
					nNextIdx = tEditLink.iLim
				end

				bFound = true
				break
			end
		end

		if bFound == false then
			strReplacedText = strReplacedText .. strText:sub(nCurrentIdx, nCurrentIdx)
		end

		nCurrentIdx = nNextIdx
	end

	return strReplacedText
end
function ForgottenChat:HelperGetChannelRef(channelType)
    for index, iterChannel in pairs(ChatSystemLib:GetChannels()) do
        if iterChannel:GetType()==channelType then
            return iterChannel
        end
    end
    return nil
end
function ForgottenChat:HelperFindRootFrame(sourceFrame)
    local sourceFrameRollup=sourceFrame
    while sourceFrameRollup:GetParent()~=nil do
		sourceFrameRollup=sourceFrameRollup:GetParent() 
	end
    return sourceFrameRollup
end
function ForgottenChat:HelperSendOutgoingMessage(conversationFrame, strText)
    local messageApproved=true
    local messageInput=conversationFrame:FindDescendent("MessageInput")
    if strText=="" then
        messageApproved=False
    elseif string.sub(strText,1,1)=="/" then
        messageApproved=False
        ChatSystemLib.Command(strText)
    end
    
    if messageApproved==true then
        strText = self:HelperReplaceLinks(strText, messageInput:GetAllLinks())
        local tInput = ChatSystemLib.SplitInput(strText)
        for index, messagePart in pairs(self:HelperSplitOutgoing(tInput.strMessage)) do
            local strMessage=""
            if conversationFrame:GetData().channel:GetType()==ChatSystemLib.ChatChannel_AccountWhisper then
                if conversationFrame:GetData().recipientRealm == "" then
                    strMessage = conversationFrame:GetData().recipientName.." "..messagePart
                else
                    strMessage = conversationFrame:GetData().recipientName.."@"..conversationFrame:GetData().recipientRealm.." "..messagePart
                end
            else
                strMessage=conversationFrame:GetData().recipientName.." "..messagePart
            end
            conversationFrame:GetData().channel:Send(strMessage)
        end
    end
    messageInput:SetText("")
    messageInput:AddHistoryString(strText)
    if Apollo.IsShiftKeyDown() == true then
        messageInput:SetFocus()
    else
        messageInput:ClearFocus()
    end
end
function ForgottenChat:HelperSendMessageFailNotice(channel, conversationFrame)
    local newLine=Apollo.LoadForm(self.xmlDoc, "ForgottenChatMessageFail", conversationFrame:FindDescendent("MessageList"), self)
    local newLineContent=XmlDoc.new()
    newLineContent:AddLine("Your last message to this user did not send properly \n \n ", ApolloColor.new("red"), "CRB_InterfaceMedium", "Center")
    newLine:SetDoc(newLineContent)
    newLine:SetHeightToContentHeight()
    conversationFrame:FindDescendent("MessageList"):ArrangeChildrenVert()
    conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())
    local frameData=conversationFrame:GetData()
    frameData.lastActivity.isSelf=nil
    conversationFrame:SetData(frameData)
end
function ForgottenChat:HelperSplitEmoticons(text)
    if trim(text)=="" then return {} end
    if self.persistantSettings.emoticonsEnabled == false then return {{sType='text',sValue=text}} end
    local split_list={}
    for emoticon, spriteName in pairs(emoticonLookup) do
        local emoticon_start, emoticon_end=string.find(text, emoticon, 1, true)
        if emoticon_start ~= nil then
            for index, segment in pairs(self:HelperSplitEmoticons(string.sub(text,1, emoticon_start-1))) do
                table.insert(split_list, segment)
            end
            table.insert(split_list,{sType="emote", sValue=emoticonLookup[emoticon]})
            for index, segment in pairs(self:HelperSplitEmoticons(string.sub(text,emoticon_end+1, string.len(text)))) do
                table.insert(split_list, segment)
            end
            return split_list
        end
    end
    return {{sType='text',sValue=text}}
end
function ForgottenChat:HelperSplitURLs(text)
    if trim(text)=="" then return {} end
    local split_list={}
    local url=self:HelperFindURL(text)
    if url~=nil then
        local url_start, url_end=string.find(text, url, 1, true)
        if url_start ~= nil then
            for index, segment in pairs(self:HelperSplitURLs(string.sub(text,1, url_start-1))) do
                table.insert(split_list, segment)
            end
            table.insert(split_list,{sType="url", sValue=url})
            for index, segment in pairs(self:HelperSplitURLs(string.sub(text,url_end+1, string.len(text)))) do
                table.insert(split_list, segment)
            end
            return split_list
        end
    end
    return {{sType='text',sValue=text}}
end
function ForgottenChat:HelperSplitOutgoing(message)
    local messageList={}
    local chunkSize=496
    if string.len(message)>chunkSize+3 then
        while string.len(message)>0 do
            table.insert(messageList, string.sub(message,1,chunkSize).."...")
            message=string.sub(message,chunkSize+1,string.len(message))
        end
        return messageList
    end
    return {message}
end
function ForgottenChat:HelperGetHexColor(color)
    local color=trim(string.lower(color))
    if string.sub(color,1,1)=="#" then color=string.sub(color, 2, string.len(color)) end
    if string.len(color)==6 and string.find(color, "^[abcdef1234567890]+$")~=nil then 
        return "ff"..color 
    end
    return color
end
function ForgottenChat:HelperSetConversationUnityMode(conversationFrame)
    if conversationFrame==nil then return end
    local unityLeft=conversationFrame:FindDescendent("UnityControlLeft")
    local unityRight=conversationFrame:FindDescendent("UnityControlRight")
    local unityNameLeft=conversationFrame:FindDescendent("UnityNameLeft")
    local unityNameRight=conversationFrame:FindDescendent("UnityNameRight")
    local unityDecorLeft=conversationFrame:FindDescendent("UnityDecorLeft")
    local unityDecorRight=conversationFrame:FindDescendent("UnityDecorRight")
    
    if self.persistantSettings.unifiedWindows==true and self:HelperCountConversations()>1 then
        if unityLeft ~= nil then unityLeft:Show(true) end
        if unityRight ~= nil then unityRight:Show(true) end
        if unityNameLeft ~= nil then 
            unityNameLeft:Show(true) 
            unityNameLeft:SetText(self:UnityGetPrev(conversationFrame):FindDescendent("UserField"):GetText())
        end
        if unityNameRight ~= nil then 
            unityNameRight:Show(true) 
            unityNameRight:SetText(self:UnityGetNext(conversationFrame):FindDescendent("UserField"):GetText())
        end
        if unityDecorLeft ~= nil then unityDecorLeft:Show(true) end
        if unityDecorRight ~= nil then unityDecorRight:Show(true) end
    else
        if unityLeft ~= nil then unityLeft:Show(false) end
        if unityRight ~= nil then unityRight:Show(false) end
        if unityNameLeft ~= nil then unityNameLeft:Show(false) end
        if unityNameRight ~= nil then unityNameRight:Show(false) end
        if unityDecorLeft ~= nil then unityDecorLeft:Show(false) end
        if unityDecorRight ~= nil then unityDecorRight:Show(false) end
    end
end
function ForgottenChat:HelperCountConversations()
    count=0
    for index, entry in pairs(self.conversations) do
        count=count+1
    end
    return count
end
function ForgottenChat:HelperGetHistoryChronoOrder()
    local returnList={}
    local newest=self:HelperGetNewestHistoryConversation(GameLib:GetLocalTime())
    table.insert(returnList, newest)
    while newest ~= nil do
        newest=self:HelperGetNewestHistoryConversation(self.persistantSettings.conversationHistory[newest].lastEvent)
        table.insert(returnList, newest)
    end
    return returnList
end
function ForgottenChat:HelperGetNewestHistoryConversation(newestBefore)
    local newestPlayerCode=nil
    for playerCode, conversationData in pairs(self.persistantSettings.conversationHistory) do
        if conversationData.lastEvent~=nil then
            if newestPlayerCode == nil and isBefore(conversationData.lastEvent, newestBefore) then
                newestPlayerCode=playerCode
            elseif newestPlayerCode ~= nil and self.persistantSettings.conversationHistory[newestPlayerCode].lastEvent~=nil then
                if isBefore(self.persistantSettings.conversationHistory[newestPlayerCode].lastEvent, conversationData.lastEvent) and isBefore(conversationData.lastEvent, newestBefore) then
                    newestPlayerCode=playerCode
                end
            end
        end
    end
    return newestPlayerCode
end
function ForgottenChat:UnityModeGoToConversation(conversationFrame)
    if self.persistantSettings.unifiedWindows==false then return end
    
    local lastTop, lastLeft, lastRight, lastBottom
    local lastMinimized, lastMinimizedHeight
    local lastCombatTicks, lastCombatClosed
    for index, conversation in pairs(self.conversations) do
        if conversation:IsShown() and conversation ~= conversationFrame then
            lastTop, lastLeft, lastRight, lastBottom=conversation:GetAnchorOffsets()
            lastMinimized=conversation:GetData().minimized
            lastMinimizedHeight=conversation:GetData().minimizedHeight
            lastCombatTicks=conversation:GetData().combatTicks
            conversation:GetData().combatTicks=nil
            lastCombatClosed=conversation:GetData().combatClosed
            conversation:GetData().combatClosed=nil
            self:HelperHideConversation(conversation)
        end
    end
    if lastMinimized~=nil and lastTop~=nil and lastLeft~=nil and lastRight~=nil and lastBottom ~=nil then
        conversationFrame:SetAnchorOffsets(lastTop, lastLeft, lastRight, lastBottom)
        if lastMinimized==true then
            self:HelperMinimizeConversation(conversationFrame)
            conversationFrame:GetData().minimizedHeight=lastMinimizedHeight
        else
            self:HelperMaximizeConversation(conversationFrame)
        end
    end
    self:UnityShowConversation(conversationFrame)
    
    conversationFrame:GetData().combatClosed=lastCombatClosed
    conversationFrame:GetData().combatTicks=lastCombatTicks
end
function ForgottenChat:UnityShowConversation(conversationFrame, combat)
    ForgottenChat:HelperShowConversation(conversationFrame, combat, true)
end
function ForgottenChat:UnityGetNext(conversationFrame)
    local useNext=false
    local nextConversation=nil
    local firstConversation=nil
    for index, conversation in pairs(self.conversations) do
        if firstConversation==nil then
            firstConversation=conversation
        end
        
        if useNext==true then
            nextConversation=conversation
            break
        elseif conversation == conversationFrame then
            useNext=true
        end
    end
    if nextConversation==nil then
        --for loop executed without finding the next one.  Use first item
        nextConversation=firstConversation
    end
    return nextConversation
end
function ForgottenChat:UnityGetPrev(conversationFrame)
    if self:HelperCountConversations() <= 1 then return end
    local previousConversation=nil
    local finalConversation=nil
    local isFound=false
    for index, conversation in pairs(self.conversations) do
        finalConversation=conversation
        if conversation == conversationFrame then
            isFound=true
        end
        if isFound == false then
            previousConversation=conversation
        end
    end
    if previousConversation==nil then
        previousConversation=finalConversation
    end
    return previousConversation
end

---------------------------------------------------------------------------------------------------
-- ForgottenChatWindow Event Functions
---------------------------------------------------------------------------------------------------
function ForgottenChat:OnConversationEscape( wndHandler, wndControl )
    --This, it does NOTHING
end
function ForgottenChat:OnConversationClose( wndHandler, wndControl, eMouseButton )
    if self.persistantSettings.unifiedWindows then
        if Apollo.IsShiftKeyDown() == true then
            self:HelperCloseAllConversations()
            return
        end
        nextConversation=self:UnityGetNext(wndHandler:GetFrame())
        self:OnConversationUnityNext(wndHandler, wndControl, eMouseButton)
        self:HelperCloseConversation(self:HelperFindRootFrame(wndHandler))
        --When we call the OnConversationUnityNext function, it does a count to see how many windows
        --are active, and if only one is active it will hide the unity interface on that window
        --However, here when we call it, the conversation being closed gets included in that count because
        --we hadn't called the close function beforehand.  Therefore, we have to manually reset the unity mode
        --We don't call the HelperCloseConversation beforehand, otherwise, it is unable to determine the
        --appropriate conversation to go to
        self:HelperSetConversationUnityMode(nextConversation)
    else
        self:HelperCloseConversation(self:HelperFindRootFrame(wndHandler))
    end
    
end
function ForgottenChat:OnConversationHeaderClick( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
    --This event gets propogated several times for each parent, but we only want one.
    if wndHandler:GetName()~=wndControl:GetName() then return end
    local conversationFrame=self:HelperFindRootFrame(wndHandler)
    conversationFrame:ToFront()
    if eMouseButton==1 then
        local actionPanel=conversationFrame:FindDescendent("ActionContainer")
        actionPanel:Show(not actionPanel:IsShown())
    elseif bDoubleClick == true then
        self:HelperToggleMinMaxConversation(conversationFrame)
        if conversationFrame:GetData().minimized~=true and self.persistantSettings.focusOnMaximize then
            conversationFrame:FindDescendent("MessageInput"):SetFocus()
        end
    end
end
function ForgottenChat:OnConversationInputReturn( wndHandler, wndControl, strText )
    self:HelperSendOutgoingMessage(self:HelperFindRootFrame(wndHandler), strText)
end
function ForgottenChat:OnConversationSendClick( wndHandler, wndControl, eMouseButton )
	self:HelperSendOutgoingMessage(self:HelperFindRootFrame(wndHandler), self:HelperFindRootFrame(wndHandler):FindDescendent("MessageInput"):GetText())
end
function ForgottenChat:OnConversationAction( wndHandler, wndControl, eMouseButton )
    local actionCode=wndControl:GetData()
    local conversationFrame=self:HelperFindRootFrame(wndControl)
    local targetName=conversationFrame:GetData().recipientName
    local targetRealm=conversationFrame:GetData().recipientRealm
    if targetRealm=="" then
        targetRealm=GameLib:GetRealmName()
    end
    if actionCode=="target" then
        GameLib.SetTargetUnit(GameLib.GetPlayerUnitByName(targetName))
    elseif actionCode=="invite" then
        GroupLib.Invite(targetName)
    elseif actionCode=="ginvite" then
        for key, guildData in pairs(GuildLib.GetGuilds()) do
            if guildData:GetType() == GuildLib.GuildType_Guild then
                guildData:Invite(targetName)
            end
        end
    elseif actionCode=="mail" then
        --
    elseif actionCode=="friend" then
        FriendshipLib.AddByName(FriendshipLib.CharacterFriendshipType_Friend, targetName, targetRealm, "Added by ForgottenChat")
    elseif actionCode=="afriend" then
        FriendshipLib.AddByName(FriendshipLib.CharacterFriendshipType_Account, targetName, targetRealm, "Added by ForgottenChat")
    elseif actionCode=="neighbor" then
        HousingLib.NeighborInviteByName(targetName)
    elseif actionCode=="ignore" then
        FriendshipLib.AddByName(FriendshipLib.CharacterFriendshipType_Ignore, targetName)
        self:HelperCloseConversation(conversationFrame)
        return
    end
    conversationFrame:FindDescendent("ActionContainer"):Show(false)
end
function ForgottenChat:OnConversationNodeClicked(wndHandler, wndControl, strNode, tAttributes, eMouseButton)
    ---------------------------------------------------------------------------------------------------
    --    This function is responsible for capturing mouse click events on nodes within the chat lines
    -- on the ForgottenChat conversation windows.  
    --
    --    The majority of the code in this function was borrowed from Carbine's default ChatLog addon, 
    -- as well as the ForgottenChatTooltip form defined in the XML.  I have slightly modified the code
    -- to allow it to work with my chat frames.
    ---------------------------------------------------------------------------------------------------

	--I am only interested in the nodes that happen to be links to items, quests or archive records
    if strNode == "Text" then
        if Apollo.IsShiftKeyDown() then
            self:HelperPrepCopyText(tAttributes.strText)
        else
            if tAttributes.strURL~=nil then
                self:HelperPrepCopyText(tAttributes.strURL)
            end
        end
	elseif strNode == "Link" then
		local nIndex = tonumber(tAttributes.strIndex)
		--When the link was added to the chat window, we recorded the relevent data in a dictionary
		--for later retrieval.  We have to make sure that the link index we are interested in still 
		--exists within the dictionary, as only a small portion of the data is retrievable from the
		--click event
		if self.tLinks[nIndex] and ( self.tLinks[nIndex].uItem or self.tLinks[nIndex].uQuest or self.tLinks[nIndex].uArchiveArticle ) then
			if Apollo.IsShiftKeyDown() == true then
                local currentMessageInput = self:HelperGetActiveInputBox()
                local tLink=self.tLinks[nIndex]
				-- pump link to the chat line
				if currentMessageInput == nil then
                    if self.coreChatAddon ~= nil then
                        self.coreChatAddon:OnNodeClick(nil,nil, strNode, tAttributes, eMouseButton)
                    end
                else
					currentMessageInput:AddLink( currentMessageInput:GetSel().cpCaret, tLink.strText, tLink )
				end
            else
				if self.tLinks[nIndex].uItem then
                    if Apollo:IsControlKeyDown() then
                        --Triggers the item to be shown in the dressing room
                        Event_FireGenericEvent("ShowItemInDressingRoom", self.tLinks[nIndex].uItem)
                    else
                        local bWindowExists = false
                        for idx, wndCur in pairs(self.twndItemLinkTooltips or {}) do
                            if wndCur:GetData() == self.tLinks[nIndex].uItem then
                                bWindowExists = true
                                break
                            end
                        end
                        if bWindowExists == false then
                            local wndChatItemToolTip = Apollo.LoadForm("ForgottenChat.xml", "ForgottenChatTooltip", nil, self)
                            wndChatItemToolTip:SetData(self.tLinks[nIndex].uItem)
                            
                            table.insert(self.twndItemLinkTooltips, wndChatItemToolTip)
                            
                            local itemEquipped = self.tLinks[nIndex].uItem:GetEquippedItemForItemType()
                            
                            local wndLink = Tooltip.GetItemTooltipForm(self, wndControl, self.tLinks[nIndex].uItem, {bPermanent = true, wndParent = wndChatItemToolTip, bSelling = false, bNotEquipped = true})
                            
                            local nLeftWnd, nTopWnd, nRightWnd, nBottomWnd = wndChatItemToolTip:GetAnchorOffsets()
                            local nLeft, nTop, nRight, nBottom = wndLink:GetAnchorOffsets()
                            
                            wndChatItemToolTip:SetAnchorOffsets(nLeftWnd, nTopWnd, nLeftWnd + nRight + 15, nBottom + 75)
                            
                            if itemEquipped then
                                wndChatItemToolTip:SetTooltipDoc(nil)
                                Tooltip.GetItemTooltipForm(self, wndChatItemToolTip, itemEquipped, {bPrimary = true, bSelling = false, bNotEquipped = false})
                            end
                        end
                    end
				elseif self.tLinks[nIndex].uQuest then
					Event_FireGenericEvent("ShowQuestLog", wndHandler:GetData()) -- Codex (todo: deprecate this)
					Event_FireGenericEvent("GenericEvent_ShowQuestLog", self.tLinks[nIndex].uQuest)
				elseif self.tLinks[nIndex].uArchiveArticle then
					Event_FireGenericEvent("HudAlert_ToggleLoreWindow")
					Event_FireGenericEvent("GenericEvent_ShowGalacticArchive", self.tLinks[nIndex].uArchiveArticle)
				end
			end
		end
	end
	return false
end
function ForgottenChat:OnConversationResize( wndHandler, wndControl )
	--I don't know what the fuck is happening here, but it was causing exceptions and stack overflows.  Maybe another day
end
function ForgottenChat:OnConversationInputGainFocus( wndHandler, wndControl)
	self:HelperFindRootFrame(wndHandler):GetData().hasFocus=true
end
function ForgottenChat:OnConversationInputLoseFocus( wndHandler, wndControl )
	self:HelperFindRootFrame(wndHandler):GetData().hasFocus=false
end
function ForgottenChat:OnConversationUnityNext( wndHandler, wndControl, eMouseButton )
    if self:HelperCountConversations() <= 1 then return end
    self:UnityModeGoToConversation(self:UnityGetNext(self:HelperFindRootFrame(wndHandler)))
end
function ForgottenChat:OnConversationUnityPrev( wndHandler, wndControl, eMouseButton )
    if self:HelperCountConversations() <= 1 then return end
    self:UnityModeGoToConversation(self:UnityGetPrev(self:HelperFindRootFrame(wndHandler)))
end
---------------------------------------------------------------------------------------------------
-- ForgottenChatTooltip Functions
---------------------------------------------------------------------------------------------------

function ForgottenChat:CloseTooltip( wndHandler, wndControl, eMouseButton )
    --Copied from Carbine's ChatLog addon, in a largely unmodified form
    local wndParent = wndControl:GetParent()
	local itemData = wndParent:GetData()
	
	for idx, wndCur in pairs(self.twndItemLinkTooltips) do
		if wndCur:GetData() == itemData then
			table.remove(self.twndItemLinkTooltips, idx)
		end
	end
	
	wndParent:Destroy()
end

---------------------------------------------------------------------------------------------------
-- ForgottenChatTextCopy Functions
---------------------------------------------------------------------------------------------------

function ForgottenChat:CloseCopyDialog( wndHandler, wndControl, eMouseButton )
    wndHandler:GetParent():Close()
end

---------------------------------------------------------------------------------------------------
-- ForgottenChatConfig Functions
---------------------------------------------------------------------------------------------------
function ForgottenChat:BuildConfigFrame()
    self.configFrame:Show(true)
    
    local configScrollableArea = self.configFrame:FindChild("ScrollableArea")
    configScrollableArea:DestroyChildren()
    --Appearance
    local appearanceHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    appearanceHeader:SetText("Appearance Settings")
    --
    local opacitySlider=Apollo.LoadForm(self.xmlDoc, "ConfigSliderTemplate", configScrollableArea, self)
    opacitySlider:FindChild("label"):SetText("Opacity")
    opacitySlider:FindChild("tooltip"):SetTooltip("This controls to what extent the main chat windows are see-through.  Fully transparent (0% opacity) windows will not be visible, but will still remain clickable.")
    opacitySlider:FindChild("slider"):SetValue(self.persistantSettings.opacity*100)
    opacitySlider:FindChild("slider"):SetData("opacity")
    self:ConfigSliderUpdate( nil, opacitySlider:FindChild("slider"), self.persistantSettings.opacity*100, nil, nil )
    --
    local skinSelect=Apollo.LoadForm(self.xmlDoc, "ConfigSelectTemplate", configScrollableArea, self)
    skinSelect:FindChild("label"):SetText("Skin")
    skinSelect:FindChild("tooltip"):SetTooltip("Choose from a selection of visual styles of your conversation windows")
    local skinSelectOptions={}
    for index, skinDetails in pairs(self.skins) do
        if skinDetails.enabled then
            skinSelectOptions[index]=skinDetails.name
        end
    end
    self:BuildConfigSelectSubmenu(skinSelect:FindChild("select"), skinSelectOptions)
    skinSelect:FindChild("select"):SetText(self.skins[self.persistantSettings.skin].name)
    skinSelect:FindChild("select"):SetData("skin")
    --
    local incomingColorInput=Apollo.LoadForm(self.xmlDoc, "ConfigTextTemplate", configScrollableArea, self)
    incomingColorInput:FindChild("label"):SetText("Incoming Text Color")
    incomingColorInput:FindChild("input"):SetData("incomingColor")
    incomingColorInput:FindChild("input"):SetText(self.persistantSettings.incomingColor)
    incomingColorInput:FindChild("indicator"):SetText("Test")
    incomingColorInput:FindChild("tooltip"):SetTooltip("Set the color of the messages you send.\n \nYou can set a 6 digit Hex color code with or without the #.\n \nYou can set an 8 digit Hex color code where the first two digits denote transparency\n \nYou can set any of the XKCD colors found at http://xkcd.com/color/rgb/ using the format 'xkcdColorName'.  For example:\nxkcdPurplishBlue\nxkcdGolden\nxkcdLightBlueGrey.")
    self:ConfigTextUpdate( nil, incomingColorInput:FindChild("input"), self.persistantSettings.incomingColor, nil, nil )
    --
    local outgoingColorInput=Apollo.LoadForm(self.xmlDoc, "ConfigTextTemplate", configScrollableArea, self)
    outgoingColorInput:FindChild("label"):SetText("Outgoing Text Color")
    outgoingColorInput:FindChild("input"):SetData("outgoingColor")
    outgoingColorInput:FindChild("input"):SetText(self.persistantSettings.outgoingColor)
    outgoingColorInput:FindChild("indicator"):SetText("Test")
    outgoingColorInput:FindChild("tooltip"):SetTooltip("Set the color of the messages you receive from others.\n \nYou can set a 6 digit Hex color code with or without the #.\n \nYou can set an 8 digit Hex color code where the first two digits denote transparency\n \nYou can set any of the XKCD colors found at http://xkcd.com/color/rgb/ using the format 'xkcdColorName'.  For example:\nxkcdPurplishBlue\nxkcdGolden\nxkcdLightBlueGrey")
    self:ConfigTextUpdate( nil, outgoingColorInput:FindChild("input"), self.persistantSettings.outgoingColor, nil, nil )
    --
    local timestampInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    timestampInput:FindChild("label"):SetText("Timestamps")
    timestampInput:FindChild("toggle"):SetData("timestamps")
    timestampInput:FindChild("toggle"):SetCheck(self.persistantSettings.timestamps)
    timestampInput:FindChild("tooltip"):Show(false)
    self:ConfigBooleanUpdate( nil, timestampInput:FindChild("toggle") )
    --
    local emoticonEnabledInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    emoticonEnabledInput:FindChild("label"):SetText("Emoticons")
    emoticonEnabledInput:FindChild("toggle"):SetData("emoticonsEnabled")
    emoticonEnabledInput:FindChild("toggle"):SetCheck(self.persistantSettings.emoticonsEnabled)
    emoticonEnabledInput:FindChild("tooltip"):Show(false)
    --
    local emoticonSelect=Apollo.LoadForm(self.xmlDoc, "ConfigSelectTemplate", configScrollableArea, self)
    emoticonSelect:FindChild("label"):SetText("Emoticon Style")
    local emoticonSelectOptions={}
    for index, emoticonDetails in pairs(self.emoticons) do
        if emoticonDetails.enabled then
            emoticonSelectOptions[index]=emoticonDetails.name
        end
    end
    self:BuildConfigSelectSubmenu(emoticonSelect:FindChild("select"), emoticonSelectOptions)
    emoticonSelect:FindChild("select"):SetText(self.emoticons[self.persistantSettings.emoticons].name)
    emoticonSelect:FindChild("select"):SetData("emoticons")
    emoticonSelect:FindChild("tooltip"):Show(false)
    --This last line has to be executed after the Emoticon Style config elements are created
    self:ConfigBooleanUpdate( nil, emoticonEnabledInput:FindChild("toggle") )
    
    --Combat
    local combatHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    combatHeader:SetText("Combat Settings")
    --
    local combatEnabledInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    combatEnabledInput:FindChild("label"):SetText("Combat Awareness")
    combatEnabledInput:FindChild("toggle"):SetData("combatEnabled")
    combatEnabledInput:FindChild("toggle"):SetCheck(self.persistantSettings.combatEnabled)
    combatEnabledInput:FindChild("tooltip"):SetTooltip("Conversations will automatically minimize or hide upon entering combat if not already in the desired state.\n \nOnly conversations automatically hidden or minimized this way will change back upon leaving combat.\n \nConversations that have focus will delay changing state until they lose focus.\n \nManually changing a conversation's state while in combat will take precedence over normal automation behavior")
    self:ConfigBooleanUpdate( nil, combatEnabledInput:FindChild("toggle") )
    --
    local combatActionSelect=Apollo.LoadForm(self.xmlDoc, "ConfigSelectTemplate", configScrollableArea, self)
    combatActionSelect:FindChild("label"):SetText("Combat Action")
    combatActionSelect:FindChild("tooltip"):SetTooltip("Choose what action is to be peformed by the combat automation upon entering combat.")
    self:BuildConfigSelectSubmenu(combatActionSelect:FindChild("select"), self.combatActions)
    combatActionSelect:FindChild("select"):SetText(self.combatActions[self.persistantSettings.combatAction])
    combatActionSelect:FindChild("select"):SetData("combatAction")
    --
    local combatHideDelaySlider=Apollo.LoadForm(self.xmlDoc, "ConfigSliderTemplate", configScrollableArea, self)
    combatHideDelaySlider:FindChild("label"):SetText("Auto Hide Delay")
    combatHideDelaySlider:FindChild("slider"):SetData("combatHideDelay")
    combatHideDelaySlider:FindChild("slider"):SetMinMax(0,30)
    combatHideDelaySlider:FindChild("slider"):SetValue(self.persistantSettings.combatHideDelay)
    combatHideDelaySlider:FindChild("tooltip"):SetTooltip("Determines how long after registering a message while in combat, a conversation will delay before performing the combat action again.\n \nNote that unless the 'Maximize On Incoming' setting is enabled, new messages will not auto-maximize even temporarily.\n \nSetting this to 0 seconds will prevent them from showing/maximizing until after combat.")
    self:ConfigSliderUpdate( nil, combatHideDelaySlider:FindChild("slider"), self.persistantSettings.combatHideDelay, nil, nil )
    --
    local combatMessageInput=Apollo.LoadForm(self.xmlDoc, "ConfigTextTemplate", configScrollableArea, self)
    combatMessageInput:FindChild("label"):SetText("Auto Reply Message")
    combatMessageInput:FindChild("input"):SetData("combatMessage")
    combatMessageInput:FindChild("input"):SetText(self.persistantSettings.combatMessage)
    combatMessageInput:FindChild("tooltip"):SetTooltip("If not blank, you will automatically respond to any incoming whispers received while in combat with this message.")
    self:ConfigTextUpdate( nil, combatMessageInput:FindChild("input"), self.persistantSettings.combatMessage, nil, nil )
    --This last line has to be executed after other other config values are defined
    self:ConfigBooleanUpdate( nil, combatEnabledInput:FindChild("toggle") )
    
    
    --History
    local historyHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    historyHeader:SetText("History Settings")
    --
    local historyLengthSlider=Apollo.LoadForm(self.xmlDoc, "ConfigSliderTemplate", configScrollableArea, self)
    historyLengthSlider:FindChild("label"):SetText("Long-Term History Length")
    historyLengthSlider:FindChild("slider"):SetData("historyLength")
    historyLengthSlider:FindChild("slider"):SetValue(self.persistantSettings.historyLength/10)
    historyLengthSlider:FindChild("tooltip"):SetTooltip("The number of messages from each conversation to retain in long term history logs.")
    self:ConfigSliderUpdate( nil, historyLengthSlider:FindChild("slider"), self.persistantSettings.historyLength/10, nil, nil )
    --
    local historyShownSlider=Apollo.LoadForm(self.xmlDoc, "ConfigSliderTemplate", configScrollableArea, self)
    historyShownSlider:FindChild("label"):SetText("Recent Summary Length")
    historyShownSlider:FindChild("slider"):SetData("historyShown")
    historyShownSlider:FindChild("slider"):SetValue(self.persistantSettings.historyShown*4)
    historyShownSlider:FindChild("tooltip"):SetTooltip("The number of lines of history to show in new conversation windows")
    self:ConfigSliderUpdate( nil, historyShownSlider:FindChild("slider"), self.persistantSettings.historyShown*4, nil, nil )
    --
    local historyViewTrigger=Apollo.LoadForm(self.xmlDoc, "ConfigTriggerTemplate", configScrollableArea, self)
    historyViewTrigger:FindChild("label"):SetText("Long Term History")
    historyViewTrigger:FindChild("trigger"):SetData("historyView")
    historyViewTrigger:FindChild("trigger"):SetText("Click To View")
    historyViewTrigger:FindChild("tooltip"):Show(false)
    
    --Behaviour
    local otherHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    otherHeader:SetText("Behaviour Settings")
    --
    local escapableWindowsInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    escapableWindowsInput:FindChild("label"):SetText("Close On Escape")
    escapableWindowsInput:FindChild("toggle"):SetData("escapableWindows")
    escapableWindowsInput:FindChild("toggle"):SetCheck(self.persistantSettings.escapableWindows)
    escapableWindowsInput:FindChild("tooltip"):SetTooltip("If enabled, hitting escape will close all open conversation windows")
    self:ConfigBooleanUpdate( nil, escapableWindowsInput:FindChild("toggle") )
    --
    local autoUnMinimizeInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    autoUnMinimizeInput:FindChild("label"):SetText("Maximize On Incoming")
    autoUnMinimizeInput:FindChild("toggle"):SetData("autoUnMinimize")
    autoUnMinimizeInput:FindChild("toggle"):SetCheck(self.persistantSettings.autoUnMinimize)
    autoUnMinimizeInput:FindChild("tooltip"):SetTooltip("If enabled, conversations will automatically be maximized when they receive a new incoming message")
    self:ConfigBooleanUpdate( nil, autoUnMinimizeInput:FindChild("toggle") )
    --
    local maxFocusInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    maxFocusInput:FindChild("label"):SetText("Focus On Maximize")
    maxFocusInput:FindChild("toggle"):SetData("focusOnMaximize")
    maxFocusInput:FindChild("toggle"):SetCheck(self.persistantSettings.focusOnMaximize)
    maxFocusInput:FindChild("tooltip"):SetTooltip("Automatically set focus to the input box when manually maximizing a conversation")
    self:ConfigBooleanUpdate( nil, maxFocusInput:FindChild("toggle") )
    --
    local unifiedInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    unifiedInput:FindChild("label"):SetText("Unified Windows")
    unifiedInput:FindChild("toggle"):SetData("unifiedWindows")
    unifiedInput:FindChild("toggle"):SetCheck(self.persistantSettings.unifiedWindows)
    unifiedInput:FindChild("tooltip"):SetTooltip("ALPHA FEATURE\n \nIf enabled, only one conversation will be visible at any given time, but gives you the ability to scroll through all active conversations easily.")
    self:ConfigBooleanUpdate( nil, unifiedInput:FindChild("toggle") )
    --
    local userInputFormatInput=Apollo.LoadForm(self.xmlDoc, "ConfigTextTemplate", configScrollableArea, self)
    userInputFormatInput:FindChild("label"):SetText("User Info Format")
    userInputFormatInput:FindChild("input"):SetData("unitInfoFormat")
    userInputFormatInput:FindChild("input"):SetText(self.persistantSettings.unitInfoFormat)
    userInputFormatInput:FindChild("tooltip"):SetTooltip("The format of the unit info text.  Uses the following replacement rules\n\n:level:\n:class:\n:race:\n:zone:\n:path:\n:name:\n:myname:\n\nAnd, the following in a few limited circumstances\n\n:title:\n:affil:\n:guild:\n:gender:\n\nFor a little fun, try :rand:")
    self:ConfigTextUpdate( nil, userInputFormatInput:FindChild("input"), self.persistantSettings.unitInfoFormat, nil, nil )
    --
    local whoEnabledInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    whoEnabledInput:FindChild("label"):SetText("Info Retrieval")
    whoEnabledInput:FindChild("toggle"):SetData("whoEnabled")
    whoEnabledInput:FindChild("toggle"):SetCheck(self.persistantSettings.whoEnabled)
    whoEnabledInput:FindChild("tooltip"):SetTooltip("Warning: Buggy at best, mostly non-functional.\n \nEnabling this will attempt to retrieve the user info of your conversational partner by performing an automated /who request.  If you have an addon that makes use of the /who functionality, enabling this might cause unusual behavior.\n\nIf disabled, we make some wild assumptions.")
    self:ConfigBooleanUpdate( nil, whoEnabledInput:FindChild("toggle") )
    --
    local captureWhispersInput=Apollo.LoadForm(self.xmlDoc, "ConfigBooleanTemplate", configScrollableArea, self)
    captureWhispersInput:FindChild("label"):SetText("Auto-Capture Whispers")
    captureWhispersInput:FindChild("toggle"):SetData("captureWhispers")
    captureWhispersInput:FindChild("toggle"):SetCheck(self.persistantSettings.captureWhispers)
    captureWhispersInput:FindChild("tooltip"):SetTooltip("If enabled, attempts to intercept whispers from getting to the default Chat interface.  Alternatively, disable this setting and hide whispers in your default chat windows.")
    self:ConfigBooleanUpdate( nil, captureWhispersInput:FindChild("toggle") )
    --
    local defaultLocationTrigger=Apollo.LoadForm(self.xmlDoc, "ConfigTriggerTemplate", configScrollableArea, self)
    defaultLocationTrigger:FindChild("label"):SetText("Set Default Location")
    defaultLocationTrigger:FindChild("trigger"):SetData("defaultLoc")
    defaultLocationTrigger:FindChild("trigger"):SetText("Click To Modify")
    defaultLocationTrigger:FindChild("tooltip"):SetTooltip("Set the initial position of new conversations.")
    
    
    --Other
    local otherHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    otherHeader:SetText("Other")
    --
    local settingsResetTrigger=Apollo.LoadForm(self.xmlDoc, "ConfigTriggerTemplate", configScrollableArea, self)
    settingsResetTrigger:FindChild("label"):SetText("Reset All Settings")
    settingsResetTrigger:FindChild("trigger"):SetData("settingsReset")
    settingsResetTrigger:FindChild("trigger"):SetText("Click to reset")
    settingsResetTrigger:FindChild("tooltip"):Show(false)
    
    --Attribution
    local attributionHeader=Apollo.LoadForm(self.xmlDoc, "ConfigHeaderTemplate", configScrollableArea, self)
    attributionHeader:SetText("Special Thanks")
    --
    local attributionMain=Apollo.LoadForm(self.xmlDoc, "ConfigAttributionTemplate", configScrollableArea, self)
    
    configScrollableArea:ArrangeChildrenVert()    
end
function ForgottenChat:ToggleConfigFrame(elementList, action)
    if self.configFrame==nil then return end
    local configScrollableArea = self.configFrame:FindChild("ScrollableArea")
    for index, element in pairs(configScrollableArea:GetChildren()) do
        for subIndex, subElement in pairs(element:GetChildren()) do
            for index, findElement in pairs(elementList) do
                if subElement:GetData()==findElement then
                    if action=="hide" then
                        element:Show(false)
                    else
                        element:Show(true)
                    end
                end
            end
        end 
    end
    configScrollableArea:ArrangeChildrenVert() 
end
function ForgottenChat:CloseConfig( wndHandler, wndControl, eMouseButton )
    wndHandler:GetParent():Show(false)
end
function ForgottenChat:ConfigSliderUpdate( wndHandler, wndControl, fNewValue, fOldValue, bOkToChange )
    wndControl:GetParent():FindChild("indicator"):SetText(self:SetConfig(wndControl:GetData(), math.floor(fNewValue)))
    wndControl:SetValue(fNewValue)
end
function ForgottenChat:ConfigTextUpdate( wndHandler, wndControl, strNewText, strOldText, bAllowed )
    local indicatorUpdate=self:SetConfig(wndControl:GetData(), strNewText)
    if indicatorUpdate~=nil then
        if indicatorUpdate.color~=nil then
            wndControl:GetParent():FindChild("indicator"):SetTextColor(self:HelperGetHexColor(indicatorUpdate.color))
        end
    end
end
function ForgottenChat:ConfigBooleanUpdate( wndHandler, wndControl, eMouseButton )  
    local checkedState=wndControl:IsChecked()
    local buttonText=self:SetConfig(wndControl:GetData(), checkedState)
    if buttonText~=nil then
        wndControl:SetText(buttonText)
    else
        if checkedState==true then
            wndControl:SetText("Enabled")
        else
            wndControl:SetText("Disabled")
        end
    end
end
function ForgottenChat:ConfigTriggerUpdate( wndHandler, wndControl, eMouseButton )
    local buttonText=self:SetConfig(wndControl:GetData(), nil)
    if buttonText~=nil then
        wndControl:SetText(buttonText)
    end
end
function ForgottenChat:ConfigSelectUpdate( wndHandler, wndControl )
    local selectMain = wndControl:GetParent():GetParent()
    wndControl:GetParent():Show(false)
    selectMain:SetText(wndControl:GetText())
    self:SetConfig(selectMain:GetData(), wndControl:GetData())
end
function ForgottenChat:ConfigDefaultLocationUpdate( wndHandler, wndControl, eMouseButton )
    local o1,o2,o3,o4=self.positionInd:GetAnchorOffsets()
    self.persistantSettings.defaultPosition={o1,o2,o3,o4}
    self.positionInd:Destroy()
    self.positionInd=nil
end
function ForgottenChat:BuildConfigSelectSubmenu(wndControl, items)
    local subMenu=Apollo.LoadForm(self.xmlDoc, "ConfigSelectSubMenu", wndControl, self)
    subMenu:SetAnchorPoints(0,1,1,1)
    local itemCount=0
    for index, item in pairs(items) do
        itemCount=itemCount+1
        local subMenuItem = Apollo.LoadForm(self.xmlDoc, "ConfigSelectSubMenuButton", subMenu, self)
        subMenuItem:SetText(item)
        subMenuItem:SetData(index)
    end
    subMenu:Invoke()
    subMenu:Show(false)
    --We anchor it so that it scales upwards, over elements already created in the config window. Scrolling it downwards causes unusual clipping behavior
	subMenu:SetAnchorOffsets(0,itemCount*-20, 0, 0)
    subMenu:ArrangeChildrenVert()
end
function ForgottenChat:ConfigSelectToggleSubmenu( wndHandler, wndControl, eMouseButton )
    local subMenu=wndHandler:FindChild("ConfigSelectSubMenu")
    if subMenu:IsShown() then
        subMenu:Show(false)
    else
        subMenu:Show(true)
    end
end
function ForgottenChat:SetConfig(name, value)
    if self.persistantSettings==nil then self.persistantSettings={} end
    if name=="historyLength" then
        local newValue=value*10
        self.persistantSettings[name]=newValue
        return newValue
    elseif name=="historyShown" then
        local newValue=math.floor(value/4)
        self.persistantSettings[name]=newValue
        return newValue
    elseif name=="opacity" then
        self.persistantSettings[name]=value/100.0
        for recipientKey, conversationFrame in pairs(self.conversations) do
			conversationFrame:SetOpacity(self.persistantSettings.opacity)
		end
        return value.."%"
    elseif name=="combatHideDelay" then
        self.persistantSettings[name]=value
        return value.."s"
    elseif name=="combatEnabled" then
        self.persistantSettings[name]=value
        self:OnCombatToggle(GameLib.GetPlayerUnit(), GameLib.GetPlayerUnit():IsInCombat())
        --Disable or enable certain elements
        if value==true then
            self:ToggleConfigFrame({"combatHideDelay","combatMessage","combatAction"},"show")
        else
            self:ToggleConfigFrame({"combatHideDelay","combatMessage","combatAction"},"hide")
        end
    elseif name=="historyView" then
        self:BuildHistoryFrame()
        self.historyFrame:Show(true)
        return nil
    elseif name=="defaultLoc" then
        if self.positionInd~=nil then return end
        self.positionInd=Apollo.LoadForm(self.xmlDoc, "DefaultPositionIndicator", nil, self)
        self.positionInd:SetOpacity(.5)
        self.positionInd:SetAnchorOffsets(self.persistantSettings.defaultPosition[1],self.persistantSettings.defaultPosition[2],self.persistantSettings.defaultPosition[3],self.persistantSettings.defaultPosition[4])
        return nil
    elseif name=="settingsReset" then
        for index, conversationFrame in pairs(self.conversations) do
            self:HelperCloseConversation(conversationFrame)
        end
        self:SetDefaultSettings()
        self:BuildConfigFrame()
    elseif name=="skin" then
        self.persistantSettings.skin=value
        for index, conversationFrame in pairs(self.conversations) do
            local user=conversationFrame:GetData().recipientName
            local realm=conversationFrame:GetData().recipientRealm
            local channel=conversationFrame:GetData().channel
            self:HelperCloseConversation(conversationFrame)
            local conversationFrame=self:HelperCreateConversation(user, realm, channel)
            conversationFrame:FindDescendent("MessageList"):ArrangeChildrenVert()
            conversationFrame:FindDescendent("MessageList"):SetVScrollPos(conversationFrame:FindDescendent("MessageList"):GetVScrollRange())	
        end
    elseif name=="combatAction" then
        self.persistantSettings.combatAction=value
        if GameLib.GetPlayerUnit():IsInCombat() == true then
            for index, conversationFrame in pairs(self.conversations) do
                if value==1 then
                    --Being set to "hide"
                   self:HelperMaximizeConversation(conversationFrame, true)
                   self:HelperHideConversation(conversationFrame, true)
                elseif value==2 then
                    --Being set to minimize
                   self:HelperShowConversation(conversationFrame, true) 
                   self:HelperMinimizeConversation(conversationFrame, true)
                end
            end
        end
    elseif name=="whoEnabled" then
        self.persistantSettings[name]=value
        --Disable or enable certain events
        if self.coreWhoAddon ~=nil then
            Apollo.RemoveEventHandler("WhoResponse",self.coreWhoAddon)
            if value==false then
                Apollo.RegisterEventHandler("WhoResponse", "OnWhoResponse", self.coreWhoAddon)
            end
        end
    elseif name=="timestamps" then
        self.persistantSettings[name]=value
    elseif name=="captureWhispers" then
        self.persistantSettings[name]=value
        --Disable or enable certain events
        if self.coreChatAddon ~=nil then
            if value==true then
                Apollo.RemoveEventHandler("ChatMessage",self.coreChatAddon)
            else
                Apollo.RegisterEventHandler("ChatMessage", "OnChatMessage", self.coreChatAddon)
            end
        end
    elseif name=="escapableWindows" then
        self.persistantSettings.escapableWindows=value
        for index, conversationFrame in pairs(self.conversations) do
            conversationFrame:SetStyle("Escapable",self.persistantSettings.escapableWindows)
        end
    elseif name=="autoUnMinimize" then
        self.persistantSettings.autoUnMinimize=value
    elseif name=="incomingColor" then
        self.persistantSettings.incomingColor=value
        return {["color"]=value}
    elseif name=="outgoingColor" then
        self.persistantSettings.outgoingColor=value
        return {["color"]=value}
    elseif name=="emoticonsEnabled" then
        self.persistantSettings[name]=value
        --Disable or enable certain elements
        if value==true then
            self:ToggleConfigFrame({"emoticons"},"show")
        else
            self:ToggleConfigFrame({"emoticons"},"hide")
        end
    elseif name=="unifiedWindows" then
        self.persistantSettings[name]=value
        self:HelperShowAllConversations()
    elseif name=="focusOnMaximize" then
        self.persistantSettings[name]=value
    else
        self.persistantSettings[name]=value
        return value
    end
end

---------------------------------------------------------------------------------------------------
-- ForgottenChatHistory Functions
---------------------------------------------------------------------------------------------------

function ForgottenChat:BuildHistoryFrame()
    self.historyFrame:FindChild("ResetPlayerHistory"):Enable(false)
	self:PlayerHistorySortList()
end
function ForgottenChat:CloseHistory( wndHandler, wndControl, eMouseButton )
    wndHandler:GetFrame():Show(false)
end
function ForgottenChat:PlayerHistoryButtonClick( wndHandler, wndControl, eMouseButton )
	local playerName=wndControl:GetData()
	local playerHistory=self.historyFrame:FindChild("HistoryArea")
	self.historyFrame:FindChild("PlayerName"):SetText(playerName)
    local timeStruct=self.persistantSettings.conversationHistory[playerName].lastEvent
	self.historyFrame:FindChild("LastMessage"):SetText("Last Message at "..timeStruct.strFormattedTime.." on "..daysOfTheWeek[timeStruct.nDayOfWeek].." "..monthsOfTheYear[timeStruct.nMonth].." "..timeStruct.nDay..", "..timeStruct.nYear)
	self.historyFrame:FindChild("ResetPlayerHistory"):SetData(playerName)
    self.historyFrame:FindChild("ResetPlayerHistory"):Enable(true)
	playerHistory:DestroyChildren()
	for index, message in pairs(self.persistantSettings.conversationHistory[wndControl:GetData()]["log"]) do
		srcName=playerName
		srcColor=self:HelperGetHexColor(self.persistantSettings.incomingColor)
		if message[1]==true then
			srcName="You"
			srcColor=self:HelperGetHexColor(self.persistantSettings.outgoingColor)
		end
		for lineIndex, line in pairs(split(message[2],"[^\r\n]+")) do
			local lineTrimmed=trim(line)
			if lineTrimmed~="" then
				historyLine = Apollo.LoadForm(self.xmlDoc, "HistoryLine", playerHistory, self)
				newLine=XmlDoc.new()
				newLine:AddLine("["..srcName.."]: "..lineTrimmed, ApolloColor.new(srcColor), "CRB_InterfaceMedium", textAlign)
				historyLine:SetDoc(newLine)
				historyLine:SetHeightToContentHeight()
			end
		end
	end
	playerHistory:ArrangeChildrenVert()
end
function ForgottenChat:PlayerHistorySortList(wndHandler, wndControl, eMouseButton)
    local playerList=self.historyFrame:FindChild("PlayerList")
	local playerNames={}
    playerList:FindChild("PlayerList"):DestroyChildren()
    if self.historyFrame:FindChild("AlphaSort"):IsChecked() then
        for playerCode, conversationData in pairs(self.persistantSettings.conversationHistory) do
            if table.getn(conversationData["log"]) > 0 then
                table.insert(playerNames, playerCode)
            end
        end
        table.sort(playerNames)
    else
        playerNames=self:HelperGetHistoryChronoOrder()
    end
	for i, playerName in pairs(playerNames) do
		playerButton = Apollo.LoadForm(self.xmlDoc, "HistoryPlayerButton", playerList, self)
		playerButton:SetText(playerName)
		playerButton:SetData(playerName)
	end
	playerList:ArrangeChildrenVert()
    self:PlayerHistoryClickFirstName()
end
function ForgottenChat:PlayerHistoryClickFirstName()
    local playerList=self.historyFrame:FindChild("PlayerList")
	if table.getn(playerList:GetChildren())>0 then
       local firstName=playerList:GetChildren()[1]
       firstName:SetCheck(true)
       self:PlayerHistoryButtonClick(firstName, firstName)
    end
end
function ForgottenChat:PlayerHistoryResetPlayerHistoryClick( wndHandler, wndControl, eMouseButton )
    local srcControl=wndControl
    local confirmDialog=self.historyFrame:FindChild("ConfirmDialog")
    if wndControl:GetName()=="ResetPlayerHistory" then
        confirmDialog:SetData(wndControl:GetData())
        confirmDialog:FindChild("ConfirmMessage"):SetText("Delete history log for "..wndControl:GetData().."?")
        confirmDialog:Show(true)
    elseif wndControl:GetName()=="ResetAllPlayerHistory" then
        confirmDialog:SetData(".*@.*")
        confirmDialog:FindChild("ConfirmMessage"):SetText("Delete history logs for all users?")
        confirmDialog:Show(true)
    end
end
function ForgottenChat:PlayerHistoryResetConfirmed( wndHandler, wndControl, eMouseButton)
    local confirmDialog=self.historyFrame:FindChild("ConfirmDialog")
    local playerName=confirmDialog:GetData()
    if playerName==".*@.*" then
        self.persistantSettings.conversationHistory={}
    else
        self.persistantSettings.conversationHistory[playerName]=nil
    end
    self:BuildHistoryFrame()
    self.historyFrame:FindChild("PlayerName"):SetText("")
    self.historyFrame:FindChild("LastMessage"):SetText("")
    self.historyFrame:FindChild("LastMessage"):SetData(nil)
    self.historyFrame:FindChild("HistoryArea"):DestroyChildren()
    self.historyFrame:FindChild("ResetPlayerHistory"):Enable(false)
    self:PlayerHistoryClickFirstName()
    confirmDialog:Show(false)
end
function ForgottenChat:PlayerHistoryResetCancel( wndHandler, wndControl, eMouseButton)
    self.historyFrame:FindChild("ConfirmDialog"):Show(false)
end
-----------------------------------------------------------------------------------------------
-- ForgottenChat Instance
-----------------------------------------------------------------------------------------------
local ForgottenChatInst = ForgottenChat:new()
ForgottenChatInst:Init()
